"""
Functions to manipulate geometrical data (point cloud, etc.)
"""

import numpy as np
from scipy.spatial.transform import Rotation
from scipy.spatial import Delaunay


def check_mesh(mesh_path):  
    """
    Check input format of a given mesh. Only .msh (v2 ASCII) and .med files 
    are currently supported.
    
    """
    if mesh_path[-4:] == ".msh":
        meshFile = open(mesh_path)
        lines = meshFile.readlines()
        if lines[1][0] != '2':
            raise TypeError(
                "Mesh file is not in version 2. Errors will appear when mirroring mesh along boundaries.")
        meshFile.close()
        mesh_path_update = mesh_path
        
    elif mesh_path[-4:] == ".med": # conversion from med to msh to keep groups
        import gmsh
        print("\n")
        print("Conversion from *.med to *.msh... \n")
        gmsh.initialize()
        gmsh.option.setNumber("Mesh.MshFileVersion", 2.2)
        gmsh.open(mesh_path)
        gmsh.write(mesh_path[:-4] + ".msh")
        gmsh.finalize()
        mesh_path_update = mesh_path[:-4] + ".msh"
        
    else: # conversion from med to msh to keep groups
        mesh_path_update = mesh_path
        raise Exception(
            "Not compatible file format. Try *.med or *.msh.")
    return mesh_path_update

def extract_bem_surface(grid):
    """
    Return index of nodes corresponding grouped surface.
    Extract from a bempp grid
    :param grid: 
    :return: 
    """
    vertices = grid.vertices
    elements = grid.elements
    domain_indices = grid.domain_indices
    domains = np.unique(domain_indices)
    groups = {}
    nodes = {}
    for j in domains:
        # initialise groups
        groups[str(j)] = []
        nodes[str(j)] = None

    for i in range(len(domain_indices)):
        for j in domains:
            if j == domain_indices[i]:
                groups[str(j)] += list(elements[:, i])

    for j in domains:
        nodes[str(j)] = np.unique(groups[str(j)])
    
    return nodes


def findClosestPoint(reference_array, x, y, z):
    """
    Finds the index and coordinates of the closest point in the reference array
    to the specified coordinates (x, y, z).

    Parameters
    ----------
    reference_array : numpy array, shape (Npoints, 3)
        Array of reference points with shape [Npoints, 3].

    x, y, z : float
        Coordinates (x, y, z) to find the closest point to (in the reference array).

    Returns
    -------
    closest_index : int
        Index of the closest point in the reference array.

    closest_coordinates : numpy array, shape (3,)
        Coordinates of the closest point in the reference array.
    """
    # print(reference_array.shape)
    # print(reference_array[20, 0])

    # distances = np.linalg.norm(reference_array - np.array([x, y, z]), axis=1)
    distances = np.zeros(len(reference_array))
    for p in range(len(reference_array)):
        distances[p] = np.sqrt((reference_array[p, 0] - x)**2 + (reference_array[p, 1] - y)**2 +
                            (reference_array[p, 2] - z)**2)

    closest_index = np.argmin(distances)
    closest_coordinates = reference_array[closest_index]
    return closest_index, closest_coordinates


def points_within_radius(R_ref, R_meas, radius, v_meas, Nfft):
    """
    Find points in R_meas within a specified radius of points in R_ref.

    Parameters:
        R_ref: Array of reference points of shape (Npoints_ref, 3).
        R_meas: Array of measured points of shape (Npoints_meas, 3).
        radius: Radius within which to search for points.

    Returns:
        List of indices of points in R_meas within the specified radius of each point in R_ref.
    """
    
    indices_within_radius = []
    coefficients = np.zeros([len(R_ref), Nfft], dtype=complex)

    for ind_point, ref_point in enumerate(R_ref):
        inc = 0
        # Calculate distance between ref_point and all points in R_meas
        distances = np.zeros(len(R_meas))
        for i in range(len(R_meas)):
            distances[i] = np.sqrt((R_meas[i, 0] - R_ref[ind_point, 0])**2
                                   + (R_meas[i, 1] - R_ref[ind_point, 1])**2
                                   + (R_meas[i, 2] - R_ref[ind_point, 2])**2)

        # Find indices of points within radius - calculate corresponding weight
        within_radius_indices = np.array([])
        while len(within_radius_indices) == 0:
            within_radius_indices = np.where(distances <= radius + inc)[0]
            inc += 5e-4
        weight = 1 - distances[within_radius_indices] / radius
        weighted_average = np.sum(weight)

        # Add indices to the list
        indices_within_radius.append(within_radius_indices)

        # store coeff
        coefficients[ind_point, :] = np.sum(v_meas[within_radius_indices, :] * weight[:, np.newaxis],
                                            axis=0) / weighted_average

    return indices_within_radius, coefficients.T


def recenterZero(X):
    """
    Recenter a point cloud at the origin (0, 0, 0) in Cartesian coordinates.

    Parameters
    ----------
    X : numpy array
        Array of shape (nPoints, 3) representing the point cloud.

    Returns
    -------
    Xcenter : numpy array
        Point cloud centered at the origin.

    Notes
    -----
    This function takes a point cloud represented by a numpy array 'X' of shape (nPoints, 3),
    where each row corresponds to a 3D point with Cartesian coordinates (x, y, z). The function
    calculates the mean values of the x, y, and z coordinates and recenters the entire point cloud
    such that its center is at the origin (0, 0, 0). The resulting 'Xcenter' array contains the
    recentered point cloud.
    """
    mean_x = np.mean(X[:, 0])
    mean_y = np.mean(X[:, 1])
    mean_z = np.mean(X[:, 2])
    Xcenter = np.zeros([len(X), 3])
    Xcenter[:, 0] = X[:, 0] - mean_x
    Xcenter[:, 1] = X[:, 1] - mean_y
    Xcenter[:, 2] = X[:, 2] - mean_z
    return Xcenter

def rotatePointCloud(point_cloud, rotation_angles_deg):
    """
    Rotate a point cloud around x, y, and z axes.

    Parameters
    ----------
    point_cloud : numpy array, shape (nPoints, 3)
        Point cloud coordinates in the shape (x, y, z).

    rotation_angles_deg : list, shape (3,)
        Rotation angles around x, y, and z axes in degrees.

    Returns
    -------
    rotated_point_cloud : numpy array, shape (nPoints, 3)
        Rotated point cloud coordinates after applying the rotations.

    Notes
    -----
    This function rotates a given point cloud by specified angles around the x, y, and z axes.
    The input 'point_cloud' should be a numpy array with shape (nPoints, 3), representing the
    Cartesian coordinates (x, y, z) of each point in the cloud. The 'rotation_angles_deg' list
    should contain three rotation angles in degrees corresponding to the rotations around the
    x, y, and z axes, respectively. The rotations are applied sequentially in the order: x, y, z.
    The function returns the rotated point cloud coordinates as a numpy array with the same shape
    as the input 'point_cloud'.
    """

    # Convert rotation angles from degrees to radians
    rotation_angles_deg = np.array(rotation_angles_deg)
    rotation_angles_rad = np.radians(rotation_angles_deg)

    # Create rotation object for each axis
    rotation_x = Rotation.from_euler('x', rotation_angles_rad[0], degrees=False)
    rotation_y = Rotation.from_euler('y', rotation_angles_rad[1], degrees=False)
    rotation_z = Rotation.from_euler('z', rotation_angles_rad[2], degrees=False)

    # Apply rotations to the point cloud
    rotated_point_cloud = rotation_x.apply(point_cloud)
    rotated_point_cloud = rotation_y.apply(rotated_point_cloud)
    rotated_point_cloud = rotation_z.apply(rotated_point_cloud)

    return rotated_point_cloud


def compute_circle_point_cloud_surface_area(point_cloud):
    # Ensure that the point cloud is a numpy array of shape [nPoints, 3]
    if point_cloud.shape[1] != 3:
        raise ValueError("Point cloud should have shape [nPoints, 3]")

    # Perform Delaunay triangulation
    tri = Delaunay(point_cloud)
    maxR = np.max(tri.max_bound)
    surface_area = np.pi * maxR**2
    return surface_area


## MICROPHONE ARRAY CREATION
def create_circular_array(theta, on_axis, rotation, radius, offset):
    # Initialize the points array
    points = np.zeros((len(theta), 3))

    # Convert theta to radians
    theta_rad = np.deg2rad(theta)
    plane_indices = ()

    if on_axis in ["+x", "x", "-x"]:
        plane_indices += (0,)
    elif on_axis in ["+y", "y", "-y"]:
        plane_indices += (1,)
    elif on_axis in ["+z", "z", "-z"]:
        plane_indices += (2,)

    if rotation in ["+x", "x", "-x"]:
        plane_indices += (0,)
    elif rotation in ["+y", "y", "-y"]:
        plane_indices += (1,)
    elif rotation in ["+z", "z", "-z"]:
        plane_indices += (2,)

    # Determine the sign based on `on_axis`
    sign = 1 if "-" not in on_axis else -1
    sign_rotation = 1 if "-" not in rotation else -1

    if plane_indices[0] == plane_indices[1]:
        raise Exception("Direction of rotation cannot be similar to on-axis value.")

    # Generate the circular coordinates
    points[:, plane_indices[0]] = radius * np.cos(theta_rad) * sign
    points[:, plane_indices[1]] = radius * np.sin(theta_rad) * sign_rotation

    # Apply the offsets
    for i in range(3):
        points[:, i] += offset[i]

    return points


def create_planar_array(length, width, micSpacing, plane, offset=[0, 0, 0], 
                      vert=False, mode=False):
    """
    Create a rectangular array of microphones on given plane. Place the corner on [x=0, y=0, z=0]

    Parameters
    ----------
    length : int or float
        rectangle length.
    width : int or float
        rectangle width.
    micSpacing : float
        distance from one microphone to another 
    plane : str
        plane 'xy', 'xz'. 'zy'.
    offset : list, optional
        Corner offset. The default is [0, 0, 0].

    Returns
    -------
    xmic : array
        Microphone location in a [nMic, 3] array. Compatible with 
        getMicPressure().
    L : array
        Array corresponding to the given length
    W : array
        Array corresponding to the given width

    """
    nMic_L = int(length / micSpacing)
    nMic_W = int(width / micSpacing)
    L = np.linspace(0, length, nMic_L) #np.arange(0, length+micSpacing, micSpacing)
    W = np.linspace(0, width, nMic_W)  #np.arange(0, width+micSpacing, micSpacing)
    nMic = nMic_L * nMic_W  #len(L)*len(W)
    xmic = np.zeros([nMic, 3])
    xOffset = np.ones([nMic, 3]) * offset
    # is there a better way to do it?
    if plane=='xy':
        dim1 = 0
        dim2 = 1
        dim3 = 2 
    elif plane=='yx':
        dim1 = 1
        dim2 = 0
        dim3 = 2
    elif plane=='xz':
        dim1 = 0
        dim2 = 2
        dim3 = 1
    elif plane=='zx':
        dim1 = 2
        dim2 = 0
        dim3 = 1
    elif plane=='yz':
        dim1 = 1
        dim2 = 2
        dim3 = 0
    elif plane=='zy':
        dim1 = 2
        dim2 = 1
        dim3 = 0
        
    i = 0
    for w in range(len(W)):
        for l in range(len(L)):
            xmic[i, dim1] = L[l]
            xmic[i, dim2] = W[w]
            xmic[i, dim3] = 0
            i += 1
    xmic += xOffset
    L += offset[dim1]
    W += offset[dim2]
    out = (xmic, L, W,)

    if vert is not False:
        xmic_n = filter_points(xmic, vert, mode=mode)
        L_n = L[np.isin(L, xmic_n[:, dim1])]
        W_n = W[np.isin(W, xmic_n[:, dim2])]
        out = (xmic_n, L_n, W_n)
    return out


def filter_points(points, boundary, mode='inside'):
    """
    Filter points based on their location relative to the boundary.

    Parameters:
    - microphones: NumPy array of shape (N, 3), representing points coordinates.
    - boundary: NumPy array of shape (M, 3), representing boundary coordinates.
    - mode: 'inside' or 'outside', specifying whether to keep microphones inside or outside the boundary.

    Returns:
    - filtered_points: NumPy array of shape (K, 3), where K <= N, containing the filtered microphone coordinates.
    """

    if mode not in ('inside', 'outside'):
        raise ValueError("Mode must be 'inside' or 'outside'.")

    if mode == 'inside':
        # Keep microphones inside the boundary
        condition = np.all((points[:, np.newaxis] >= boundary.min(axis=0)) &
                           (points[:, np.newaxis] <= boundary.max(axis=0)), axis=2)
    else:
        # Keep microphones outside the boundary
        condition = np.any((points[:, np.newaxis] < boundary.min(axis=0)) |
                           (points[:, np.newaxis] > boundary.max(axis=0)), axis=2)

    filtered_points = points[condition.all(axis=1)]
    return filtered_points

def create_spherical_array(nMic, sphereRadius=1.8, offset=[0, 0, 0]):
    """
    create a spherical microphone array
    :param Nmic: number of microphones in the array
    :return: xmic: cartesian coordinates of each microphone of the array
    """

    theta, phi = np.linspace(0, 2 * np.pi, int(np.sqrt(nMic))), np.linspace(0, np.pi, int(np.sqrt(nMic)))
    THETA, PHI = np.meshgrid(theta, phi)
    R = sphereRadius
    X = R * np.sin(PHI) * np.cos(THETA) + offset[0]
    Y = R * np.sin(PHI) * np.sin(THETA) + offset[1]
    Z = R * np.cos(PHI) + offset[2]

    mic = 0
    xmic = np.zeros([nMic, 3])
    for i in range(int(np.sqrt(nMic))):
        for j in range(int(np.sqrt(nMic))):
            xmic[mic, 0] = X[i, j]
            xmic[mic, 1] = Y[i, j]
            xmic[mic, 2] = Z[i, j]
            mic += 1

    return xmic


def create_bounding_box(Lx, Ly, Lz, step=1, offset=[0, 0, 0]):
    x_offset, y_offset, z_offset = offset
    x_range = np.arange(x_offset, x_offset + Lx + step, step)
    y_range = np.arange(y_offset, y_offset + Ly + step, step)
    z_range = np.arange(z_offset, z_offset + Lz + step, step)
    
    nx = len(x_range)
    ny = len(y_range)
    nz = len(z_range)
    dim = [nx, ny, nz]
    
    x_points, y_points, z_points = np.meshgrid(x_range, y_range, z_range)

    points = np.vstack([x_points.flatten(), y_points.flatten(), z_points.flatten()]).T

    return points, dim


"""Functions for creating potential operators.

Documentation of the functionality of this module and examples
can be found at https://bempp.com/handbook/api/potential_operators.html.
"""

from . import laplace
from . import helmholtz
from . import maxwell
from . import modified_helmholtz



"""
Functions to automate the creation of simple shoebox-like geometries
"""

import gmsh


class shoebox:
    def __init__(self, Lx, Ly, Lz, position="center", 
                 minSize=0.0057, maxSize=0.057):
        """
        Create a parallelepipede.

        Parameters
        ----------
        Lx : float
            Length along x axis.
        Ly : float
            Length along y axis.
        Lz : float
            Length along z axis.
        position : str, optional
            How the box is placed: "center" will put its center at (x, y, z) = 0,
            "corner" will place the lower left corner of the box at (x, y, z) = 0. The default is "center".
        meshSize : float, optional
            Mesh size in metres. The default is 0.057 (343/1000/6).

        Returns
        -------
        None.

        """
        self.Lx = Lx
        self.Ly = Ly
        self.Lz = Lz
        self.position = position
        self.minSize = minSize
        self.maxSize = maxSize
        self.membrane = []
        self.name = []


    def addCircularMembrane(self, face, x, y, radius, physical_group, name=None):
        self.membrane.append([face, x, y, radius, physical_group])
        self.name.append(name)
              
    def addRectangularMembrane(self, face, x, y, lx, ly, physical_group, name=None):
        self.membrane.append([face, x, y, lx, ly, physical_group])
        self.name.append(name)
        
    def addPolygon(self, face, X, Y, physical_group, name=None):
        self.membrane.append([face, X, Y, physical_group])
        self.name.append(name)
        
    def build(self, path=None):
        """
        Build the geometry.

        Raises
        ------
        ValueError
            Raise error if 'position' is not correctly defined ('center' or 'corner').

        Returns
        -------
        None.

        """
        Lx = self.Lx
        Ly = self.Ly
        Lz = self.Lz
        args = self.membrane
        radSurf = []
        
        gmsh.initialize()
        gmsh.option.setNumber("Mesh.CharacteristicLengthMin", self.minSize)
        gmsh.option.setNumber("Mesh.CharacteristicLengthMax", self.maxSize)
        gmsh.option.setNumber("Mesh.MshFileVersion",2.2)   

        if self.position == "center":
            point_1 = gmsh.model.geo.addPoint(Lx/2, -Ly/2, -Lz/2)
            point_2 = gmsh.model.geo.addPoint(Lx/2, Ly/2, -Lz/2)
            point_3 = gmsh.model.geo.addPoint(Lx/2, Ly/2, Lz/2)
            point_4 = gmsh.model.geo.addPoint(Lx/2, -Ly/2, Lz/2)
            point_5 = gmsh.model.geo.addPoint(-Lx/2, -Ly/2, -Lz/2)
            point_6 = gmsh.model.geo.addPoint(-Lx/2, Ly/2, -Lz/2)
            point_7 = gmsh.model.geo.addPoint(-Lx/2, Ly/2, Lz/2)
            point_8 = gmsh.model.geo.addPoint(-Lx/2, -Ly/2, Lz/2)
        elif self.position == "corner":
            point_1 = gmsh.model.geo.addPoint(Lx, 0, 0)
            point_2 = gmsh.model.geo.addPoint(Lx, Ly, 0)
            point_3 = gmsh.model.geo.addPoint(Lx, Ly, Lz)
            point_4 = gmsh.model.geo.addPoint(Lx, 0, Lz)
            point_5 = gmsh.model.geo.addPoint(0, 0, 0)
            point_6 = gmsh.model.geo.addPoint(0, Ly, 0)
            point_7 = gmsh.model.geo.addPoint(0, Ly, Lz)
            point_8 = gmsh.model.geo.addPoint(0, 0, Lz)    
        else:
            raise ValueError("'position' should either be 'center' or 'corner'.")
        
        
        ## BUILD LINES
        # panel 1 (front)
        line_1 = gmsh.model.geo.add_line(point_1, point_2)
        line_2 = gmsh.model.geo.add_line(point_2, point_3)
        line_3 = gmsh.model.geo.add_line(point_3, point_4)
        line_4 = gmsh.model.geo.add_line(point_4, point_1)
        # panel 2 (right)
        line_5 = gmsh.model.geo.add_line(point_2, point_6)
        line_6 = gmsh.model.geo.add_line(point_6, point_7)
        line_7 = gmsh.model.geo.add_line(point_7, point_3)
        line_8 = gmsh.model.geo.add_line(point_3, point_2)
        # panel 3 (top)
        line_9 = gmsh.model.geo.add_line(point_3, point_7)
        line_10 = gmsh.model.geo.add_line(point_7, point_8)
        line_11 = gmsh.model.geo.add_line(point_8, point_4)
        line_12 = gmsh.model.geo.add_line(point_4, point_3)
        # panel 4 (bottom)
        line_13 = gmsh.model.geo.add_line(point_2, point_1)
        line_14 = gmsh.model.geo.add_line(point_1, point_5)
        line_15 = gmsh.model.geo.add_line(point_5, point_6)
        line_16 = gmsh.model.geo.add_line(point_6, point_2)
        # panel 5 (left)
        line_17 = gmsh.model.geo.add_line(point_5, point_1)
        line_18 = gmsh.model.geo.add_line(point_1, point_4)
        line_19 = gmsh.model.geo.add_line(point_4, point_8)
        line_20 = gmsh.model.geo.add_line(point_8, point_5)
        # panel 6 (back)
        line_21 = gmsh.model.geo.add_line(point_7, point_6)
        line_22 = gmsh.model.geo.add_line(point_6, point_5)
        line_23 = gmsh.model.geo.add_line(point_5, point_8)
        line_24 = gmsh.model.geo.add_line(point_8, point_7)

        ## MAKE CURVE LOOP FROM LINES
        # curve loop
        loop_1 = gmsh.model.geo.addCurveLoop([line_1, line_2, line_3, line_4])          # +x
        loop_2 = gmsh.model.geo.addCurveLoop([line_5, line_6, line_7, line_8])          # +y
        loop_3 = gmsh.model.geo.addCurveLoop([line_9, line_10, line_11, line_12])       # +z
        loop_4 = gmsh.model.geo.addCurveLoop([line_13, line_14, line_15, line_16])      # -z
        loop_5 = gmsh.model.geo.addCurveLoop([line_17, line_18, line_19, line_20])      # -y
        loop_6 = gmsh.model.geo.addCurveLoop([line_21, line_22, line_23, line_24])      # -x
        
        ## LIST OF PANELS TO BUILD 
        panel_xp = [loop_1]
        panel_xm = [loop_6]
        panel_yp = [loop_2]
        panel_ym = [loop_5]
        panel_zp = [loop_3]
        panel_zm = [loop_4]
        
        ## BUILD PANELS
        for i in range(len(args)):
            piston_tmp = args[i]
            if len(piston_tmp) == 5:
                piston_type = "circle"
            elif len(piston_tmp) == 6:
                piston_type = "rectangle"
            elif len(piston_tmp) == 4:
                piston_type = "polygon"
            
            if piston_type == "circle":
                face = piston_tmp[0]
                x    = piston_tmp[1]
                y    = piston_tmp[2]
                r    = piston_tmp[3]
        
                if face in ["x", "+x", "X", "+X", "-x", "-X"]:
                    if '-' in face:
                        sign = -1
                        Lx_offset = 0
                        Ly_offset = Ly
                    else:
                        sign = 1
                        Lx_offset = Lx
                        Ly_offset = 0
                        
                    if self.position == "center":
                        point_p1 = gmsh.model.geo.addPoint(sign * Lx/2, sign*x, y)
                        point_p2 = gmsh.model.geo.addPoint(sign * Lx/2, sign*(x+r), y)
                        point_p3 = gmsh.model.geo.addPoint(sign * Lx/2, sign*x, y+r)
                        point_p4 = gmsh.model.geo.addPoint(sign * Lx/2, sign*(x-r), y)
                        point_p5 = gmsh.model.geo.addPoint(sign * Lx/2, sign*x, y-r)
                    elif self.position == "corner":
                        point_p1 = gmsh.model.geo.addPoint(Lx_offset, Ly_offset + sign*x, y)
                        point_p2 = gmsh.model.geo.addPoint(Lx_offset, Ly_offset + sign*(x+r), y)
                        point_p3 = gmsh.model.geo.addPoint(Lx_offset, Ly_offset + sign*x, y+r)
                        point_p4 = gmsh.model.geo.addPoint(Lx_offset, Ly_offset + sign*(x-r), y)
                        point_p5 = gmsh.model.geo.addPoint(Lx_offset, Ly_offset + sign*x, y-r)
                    circle_1 = gmsh.model.geo.addCircleArc(point_p2, point_p1, point_p3)
                    circle_2 = gmsh.model.geo.addCircleArc(point_p3, point_p1, point_p4)
                    circle_3 = gmsh.model.geo.addCircleArc(point_p4, point_p1, point_p5)
                    circle_4 = gmsh.model.geo.addCircleArc(point_p5, point_p1, point_p2)
                    piston_loop = gmsh.model.geo.addCurveLoop([circle_1, circle_2, 
                                                               circle_3, circle_4])
                    radSurf.append(gmsh.model.geo.addPlaneSurface([piston_loop]))
                    
                    if "-" in face:
                        panel_xm.append(piston_loop)
                    else:
                        panel_xp.append(piston_loop)
                            
                elif face in ["y", "+y", "Y", "+Y", "-y", "-Y"]:
                    if "-" in face:
                        sign = -1
                        Lx_offset = 0
                        Ly_offset = 0
                    else:
                        sign = 1
                        Lx_offset = Lx
                        Ly_offset = Ly
                        
                    if self.position == "center":
                        point_p1 = gmsh.model.geo.addPoint(-sign*x, sign*Ly/2, y)
                        point_p2 = gmsh.model.geo.addPoint(-sign*(x+r), sign*Ly/2, y)
                        point_p3 = gmsh.model.geo.addPoint(-sign*x, sign*Ly/2, y+r)
                        point_p4 = gmsh.model.geo.addPoint(-sign*(x-r), sign*Ly/2, y)
                        point_p5 = gmsh.model.geo.addPoint(-sign*x, sign*Ly/2, y-r)
                    elif self.position == "corner":
                        point_p1 = gmsh.model.geo.addPoint(Lx_offset-sign*x, Ly_offset, y)
                        point_p2 = gmsh.model.geo.addPoint(Lx_offset-sign*(x+r), Ly_offset, y)
                        point_p3 = gmsh.model.geo.addPoint(Lx_offset-sign*x, Ly_offset, y+r)
                        point_p4 = gmsh.model.geo.addPoint(Lx_offset-sign*(x-r), Ly_offset, y)
                        point_p5 = gmsh.model.geo.addPoint(Lx_offset-sign*x, Ly_offset, y-r)
                    circle_1 = gmsh.model.geo.addCircleArc(point_p2, point_p1, point_p3)
                    circle_2 = gmsh.model.geo.addCircleArc(point_p3, point_p1, point_p4)
                    circle_3 = gmsh.model.geo.addCircleArc(point_p4, point_p1, point_p5)
                    circle_4 = gmsh.model.geo.addCircleArc(point_p5, point_p1, point_p2)
                    piston_loop = gmsh.model.geo.addCurveLoop([circle_1, circle_2, 
                                                               circle_3, circle_4])
                    radSurf.append(gmsh.model.geo.addPlaneSurface([piston_loop]))
                    
                    if "-" in face:
                        panel_ym.append(piston_loop)         
                    else:
                        panel_yp.append(piston_loop)            
        
        
                elif face in ["z", "+z", "Z", "+Z", "-z", "-Z"]:
                    if "-" in face:
                        sign = -1
                        Lz_offset = 0
                        Lx_offset = Lx
                    else:
                        sign = 1
                        Lz_offset = Lz
                        Lx_offset = 0
                                        
                    if self.position == "center":
                        point_p1 = gmsh.model.geo.addPoint(sign*x, y, sign*Lz/2)
                        point_p2 = gmsh.model.geo.addPoint(sign*(x+r), y, sign*Lz/2)
                        point_p3 = gmsh.model.geo.addPoint(sign*x, (y+r), sign*Lz/2)
                        point_p4 = gmsh.model.geo.addPoint(sign*(x-r), y, sign*Lz/2)
                        point_p5 = gmsh.model.geo.addPoint(sign*x, (y-r), sign*Lz/2)
                    elif self.position == "corner":
                        point_p1 = gmsh.model.geo.addPoint(Lx_offset+sign*x, y, Lz_offset)
                        point_p2 = gmsh.model.geo.addPoint(Lx_offset+sign*(x+r), y, Lz_offset)
                        point_p3 = gmsh.model.geo.addPoint(Lx_offset+sign*x, (y+r), Lz_offset)
                        point_p4 = gmsh.model.geo.addPoint(Lx_offset+sign*(x-r), y, Lz_offset)
                        point_p5 = gmsh.model.geo.addPoint(Lx_offset+sign*x, y-r, Lz_offset)
                    circle_1 = gmsh.model.geo.addCircleArc(point_p2, point_p1, point_p3)
                    circle_2 = gmsh.model.geo.addCircleArc(point_p3, point_p1, point_p4)
                    circle_3 = gmsh.model.geo.addCircleArc(point_p4, point_p1, point_p5)
                    circle_4 = gmsh.model.geo.addCircleArc(point_p5, point_p1, point_p2)
                    piston_loop = gmsh.model.geo.addCurveLoop([circle_1, circle_2, 
                                                               circle_3, circle_4])
                    radSurf.append(gmsh.model.geo.addPlaneSurface([piston_loop]))
                    
                    if "-" in face:
                        panel_zm.append(piston_loop)         
                    else:
                        panel_zp.append(piston_loop) 
            
        
            elif piston_type == "rectangle":                
                face = piston_tmp[0]
                x    = piston_tmp[1]
                y    = piston_tmp[2]
                lx   = piston_tmp[3]
                ly   = piston_tmp[4]                
                
                if face in ["x", "+x", "X", "+X", "-x", "-X"]:
                    if '-'  in face:
                        Lxc_offset = -Lx/2
                        Lxr_offset = 0
                        Lyr_offset = Ly
                        sign = -1
                    else:
                        Lxc_offset = Lx/2
                        Lxr_offset = Lx
                        Lyr_offset = 0
                        sign = 1
                        
                    if self.position == "center":
                        point_p1 = gmsh.model.geo.addPoint(Lxc_offset, sign*(x-lx/2), y-ly/2)
                        point_p2 = gmsh.model.geo.addPoint(Lxc_offset, sign*(x+lx/2), y-ly/2)
                        point_p3 = gmsh.model.geo.addPoint(Lxc_offset, sign*(x+lx/2), y+ly/2)
                        point_p4 = gmsh.model.geo.addPoint(Lxc_offset, sign*(x-lx/2), y+ly/2)
                    elif self.position == "corner":
                        point_p1 = gmsh.model.geo.addPoint(Lxr_offset, Lyr_offset + sign*x, y)
                        point_p2 = gmsh.model.geo.addPoint(Lxr_offset, Lyr_offset + sign*(x+lx), y)
                        point_p3 = gmsh.model.geo.addPoint(Lxr_offset, Lyr_offset + sign*(x+lx), y+ly)
                        point_p4 = gmsh.model.geo.addPoint(Lxr_offset, Lyr_offset + sign*x, y+ly)
                    
                    rect_1 = gmsh.model.geo.addLine(point_p1, point_p2)
                    rect_2 = gmsh.model.geo.addLine(point_p2, point_p3)
                    rect_3 = gmsh.model.geo.addLine(point_p3, point_p4)
                    rect_4 = gmsh.model.geo.addLine(point_p4, point_p1)
                    piston_loop = gmsh.model.geo.addCurveLoop([rect_1, rect_2, 
                                                               rect_3, rect_4])
                    radSurf.append(gmsh.model.geo.addPlaneSurface([piston_loop]))
                    
                    if "-" in face:
                        panel_xm.append(piston_loop)         
                    else:
                        panel_xp.append(piston_loop)
        
                
                if face in ["y", "+y", "Y", "+Y", "-y", "-Y"]:
                    if '-' in face:
                        Lyc_offset = -Ly/2
                        Lyr_offset = 0
                        Lxr_offset = 0
                        sign = -1
                    else:
                        Lyc_offset = Ly/2
                        Lyr_offset = Ly
                        Lxr_offset = Lx
                        sign = 1
                        
                    if self.position == "center":
                        point_p1 = gmsh.model.geo.addPoint(sign*(x+lx/2), Lyc_offset, y-ly/2)
                        point_p2 = gmsh.model.geo.addPoint(sign*(x-lx/2), Lyc_offset, y-ly/2)
                        point_p3 = gmsh.model.geo.addPoint(sign*(x-lx/2), Lyc_offset, y+ly/2)
                        point_p4 = gmsh.model.geo.addPoint(sign*(x+lx/2), Lyc_offset, y+ly/2)
                    elif self.position == "corner":
                        point_p1 = gmsh.model.geo.addPoint(Lxr_offset - sign*x, Lyr_offset, y)
                        point_p2 = gmsh.model.geo.addPoint(Lxr_offset - sign*(x+lx), Lyr_offset, y)
                        point_p3 = gmsh.model.geo.addPoint(Lxr_offset - sign*(x+lx), Lyr_offset, y+ly)
                        point_p4 = gmsh.model.geo.addPoint(Lxr_offset - sign*x, Lyr_offset, y+ly)
                    
                    rect_1 = gmsh.model.geo.addLine(point_p1, point_p2)
                    rect_2 = gmsh.model.geo.addLine(point_p2, point_p3)
                    rect_3 = gmsh.model.geo.addLine(point_p3, point_p4)
                    rect_4 = gmsh.model.geo.addLine(point_p4, point_p1)
                    piston_loop = gmsh.model.geo.addCurveLoop([rect_1, rect_2, 
                                                               rect_3, rect_4])
                    radSurf.append(gmsh.model.geo.addPlaneSurface([piston_loop]))
                     
                    if "-" in face:
                        panel_ym.append(piston_loop)         
                    else:
                        panel_yp.append(piston_loop)   
                        
                if face in ["z", "+z", "Z", "+Z", "-z", "-Z"]:                    
                    if '-' in face:
                        Lzc_offset = -Lz/2
                        Lzr_offset = 0
                        Lxr_offset = Lx
                        sign = -1 
                    else:
                        Lzc_offset = Lz/2
                        Lzr_offset = Lz
                        Lxr_offset = 0
                        sign = 1
                        
                    if self.position == "center":
                        point_p1 = gmsh.model.geo.addPoint(sign*(x+lx/2), y-ly/2, Lzc_offset)
                        point_p2 = gmsh.model.geo.addPoint(sign*(x-lx/2), y-ly/2, Lzc_offset)
                        point_p3 = gmsh.model.geo.addPoint(sign*(x-lx/2), y+ly/2, Lzc_offset)
                        point_p4 = gmsh.model.geo.addPoint(sign*(x+lx/2), y+ly/2, Lzc_offset)
                        pos_sign = -1
                    elif self.position == "corner":
                        point_p1 = gmsh.model.geo.addPoint(Lxr_offset + sign*x, y, Lzr_offset)
                        point_p2 = gmsh.model.geo.addPoint(Lxr_offset + sign*(x+lx), y, Lzr_offset)
                        point_p3 = gmsh.model.geo.addPoint(Lxr_offset + sign*(x+lx), y+ly, Lzr_offset)
                        point_p4 = gmsh.model.geo.addPoint(Lxr_offset + sign*x, y+ly, Lzr_offset)
                        pos_sign = 1
                    
                    rect_1 = gmsh.model.geo.addLine(point_p1, point_p2)
                    rect_2 = gmsh.model.geo.addLine(point_p2, point_p3)
                    rect_3 = gmsh.model.geo.addLine(point_p3, point_p4)
                    rect_4 = gmsh.model.geo.addLine(point_p4, point_p1)
                    piston_loop = gmsh.model.geo.addCurveLoop([pos_sign*rect_1, pos_sign*rect_2, 
                                                               pos_sign*rect_3, pos_sign*rect_4])
                    radSurf.append(gmsh.model.geo.addPlaneSurface([piston_loop]))
                     
                    if "-" in face:
                        panel_zm.append(piston_loop)         
                    else:
                        panel_zp.append(piston_loop)   
                                
        nonRadSurf = []
        nonRadSurf.append(gmsh.model.geo.addPlaneSurface(panel_xp))
        nonRadSurf.append(gmsh.model.geo.addPlaneSurface(panel_xm))
        nonRadSurf.append(gmsh.model.geo.addPlaneSurface(panel_yp))
        nonRadSurf.append(gmsh.model.geo.addPlaneSurface(panel_ym))
        nonRadSurf.append(gmsh.model.geo.addPlaneSurface(panel_zp))
        nonRadSurf.append(gmsh.model.geo.addPlaneSurface(panel_zm))
        gmsh.model.geo.synchronize()
        
        ## SURFACE GROUP
        for i in range(len(args)):
            piston_tmp = args[i]
            if self.name[i] is None:
                name = f"rad_surf_{i+1}"
            else:
                name = self.name[i]
            gmsh.model.addPhysicalGroup(2, [radSurf[i]], tag=piston_tmp[-1], name=name)
        
        gmsh.model.addPhysicalGroup(2, nonRadSurf, name="rigid_boundary")
        
        ## BUILD GEOMETRY        
        gmsh.model.mesh.generate(dim=2)
        
        if path is None:
            gmsh.write("geo_mesh.msh")
        elif type(path) == str:
            if path[-4::] in [".med", ".msh"]:
                gmsh.write(path)
            else:
                gmsh.finalize()
                raise Exception("Mesh extension not supported. Try '.med' or '.msh'.")
            
        gmsh.finalize()
        return None
        
        


"""Interfaces to Helmholtz operators."""

import numpy as _np

from bempp_cl.api.operators.boundary import common as _common
from bempp_cl.api.assembly.boundary_operator import BoundaryOperator as _BoundaryOperator


def single_layer(
    domain,
    range_,
    dual_to_range,
    wavenumber,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the Helmholtz single-layer boundary operator."""
    from .modified_helmholtz import single_layer as _modified_single_layer

    if _np.real(wavenumber) == 0:
        return _modified_single_layer(
            domain,
            range_,
            dual_to_range,
            _np.imag(wavenumber),
            parameters,
            assembler,
            device_interface,
            precision,
        )

    return _common.create_operator(
        "helmholtz_single_layer_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [_np.real(wavenumber), _np.imag(wavenumber)],
        "helmholtz_single_layer",
        "default_scalar",
        device_interface,
        precision,
        True,
    )


def double_layer(
    domain,
    range_,
    dual_to_range,
    wavenumber,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the Helmholtz double-layer boundary operator."""
    from .modified_helmholtz import double_layer as _modified_double_layer

    if _np.real(wavenumber) == 0:
        return _modified_double_layer(
            domain,
            range_,
            dual_to_range,
            _np.imag(wavenumber),
            parameters,
            assembler,
            device_interface,
            precision,
        )

    return _common.create_operator(
        "helmholtz_double_layer_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [_np.real(wavenumber), _np.imag(wavenumber)],
        "helmholtz_double_layer",
        "default_scalar",
        device_interface,
        precision,
        True,
    )


def adjoint_double_layer(
    domain,
    range_,
    dual_to_range,
    wavenumber,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the Helmholtz adj. double-layer boundary operator."""
    from .modified_helmholtz import (
        adjoint_double_layer as _modified_adjoint_double_layer,
    )

    if _np.real(wavenumber) == 0:
        return _modified_adjoint_double_layer(
            domain,
            range_,
            dual_to_range,
            _np.imag(wavenumber),
            parameters,
            assembler,
            device_interface,
            precision,
        )

    return _common.create_operator(
        "helmholtz_adjoint_double_layer_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [_np.real(wavenumber), _np.imag(wavenumber)],
        "helmholtz_adjoint_double_layer",
        "default_scalar",
        device_interface,
        precision,
        True,
    )


def hypersingular(
    domain,
    range_,
    dual_to_range,
    wavenumber,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the Helmholtz hypersingular boundary operator."""
    from .modified_helmholtz import hypersingular as _hypersingular

    if domain.shapeset.identifier != "p1_discontinuous":
        raise ValueError("Domain shapeset must be of type 'p1_discontinuous'.")

    if dual_to_range.shapeset.identifier != "p1_discontinuous":
        raise ValueError("Dual to range shapeset must be of type 'p1_discontinuous'.")

    if _np.real(wavenumber) == 0:
        return _hypersingular(
            domain,
            range_,
            dual_to_range,
            _np.imag(wavenumber),
            parameters,
            assembler,
            device_interface,
            precision,
        )

    return _common.create_operator(
        "helmholtz_hypersingular_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [_np.real(wavenumber), _np.imag(wavenumber)],
        "helmholtz_single_layer",
        "helmholtz_hypersingular",
        device_interface,
        precision,
        True,
    )


def multitrace_operator(
    grid,
    wavenumber,
    target=None,
    space_type="p1",
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """
    Simplified version of multitrace operator assembly.

    Parameters
    ----------
    grid : Grid
        Bempp grid object.
    wavenumber : complex
        A real or complex wavenumber
    target : Grid
        The grid for the range spaces. If target is None then
        target is set to the input grid (that is the domain
        grid).
    space_type : string
        Controls which discretisation spaces are used.
        Supported values:
        - "p1", which means that all operators are discretised
        with P1 basis functions.
        - "p1-dp0", which means that P1 basis functions are
        used for the first row and column, and DP0 for the second.
        - "p1-dual0", which means that P1 basis functions are
        used for the first row and column, and DUAL0 for the second.
    parameters : Parameters
        An optional parameters object.
    assembler : string
        The assembler type.
    device_interface : DeviceInterface
        The device interface object to be used.
    precision : string
        Either "single" or "double" for single or
        double precision mode.

    Output
    ------
    The Helmholtz multitrace operator of the form
    [[-dlp, slp], [hyp, adj_dlp]], where
    dlp : double layer boundary operator
    slp : single layer boundary operator
    hyp : hypersingular boundary operator
    adj_dlp : adjoint double layer boundary operator.

    """
    import bempp_cl.api
    from bempp_cl.api.assembly.blocked_operator import BlockedOperator

    if space_type == "p1":
        space1 = bempp_cl.api.function_space(grid, "P", 1)
        space0 = space1
        if target is not None:
            target_space1 = bempp_cl.api.function_space(target, "P", 1)
            target_space0 = target_space1
            target_space_dual1 = target_space1
            target_space_dual0 = target_space1
        else:
            target_space1 = space1
            target_space0 = space1
            target_space_dual1 = space1
            target_space_dual0 = space1
    elif space_type == "p1-dp0":
        space1 = bempp_cl.api.function_space(grid, "P", 1)
        space0 = bempp_cl.api.function_space(grid, "DP", 0)
        if target is not None:
            target_space1 = bempp_cl.api.function_space(target, "P", 1)
            target_space0 = bempp_cl.api.function_space(target, "DP", 0)
            target_space_dual1 = target_space1
            target_space_dual0 = target_space0
    elif space_type == "p1-dual":
        space1 = bempp_cl.api.function_space(grid, "P", 1)
        space0 = bempp_cl.api.function_space(grid, "DUAL", 0)
        if target is not None:
            target_space1 = bempp_cl.api.function_space(target, "P", 1)
            target_space0 = bempp_cl.api.function_space(target, "DUAL", 0)
            target_space_dual1 = target_space1
            target_space_dual0 = target_space0
    else:
        raise ValueError(f"Unknown space type: {space_type}")

    slp = single_layer(
        space0,
        target_space1,
        target_space_dual0,
        wavenumber,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    dlp = double_layer(
        space1,
        target_space1,
        target_space_dual0,
        wavenumber,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    hyp = hypersingular(
        space1,
        target_space0,
        target_space_dual1,
        wavenumber,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    adj_dlp = adjoint_double_layer(
        space0,
        target_space0,
        target_space_dual1,
        wavenumber,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    blocked = BlockedOperator(2, 2)

    blocked[0, 0] = -dlp
    blocked[0, 1] = slp
    blocked[1, 0] = hyp
    blocked[1, 1] = adj_dlp

    return blocked


def osrc_dtn(
    space,
    wavenumber,
    npade=2,
    theta=_np.pi / 3.0,
    damped_wavenumber=None,
    parameters=None,
    device_interface=None,
    precision=None,
):
    """Assemble the OSRC approximation to the DtN operator."""
    if space.shapeset.identifier != "p1_discontinuous":
        raise ValueError("Space shapeset must be of type 'p1_discontinuous'.")

    return _OsrcDtN(
        space,
        parameters,
        [wavenumber, npade, theta, damped_wavenumber],
        device_interface,
        precision,
    )


class _OsrcDtN(_BoundaryOperator):
    """Implementation of the OSRC DtN operator."""

    def __init__(self, space, parameters, operator_options, device_interface=None, precision=None):
        from bempp_cl.api.operators import OperatorDescriptor

        super().__init__(space, space, space, parameters)

        self._device_interface = device_interface

        self._operator_descriptor = OperatorDescriptor(
            "osrc_dtn",
            operator_options,
            "laplace_beltrami",
            "default_sparse",
            precision,
            True,
            None,
            1,
        )

    @property
    def descriptor(self):
        """Operator descriptor."""
        return self._operator_descriptor

    def _assemble(self):
        """Assemble the operator."""
        from bempp_cl.api.operators.boundary.sparse import identity
        from bempp_cl.api.operators.boundary.sparse import laplace_beltrami
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            InverseSparseDiscreteBoundaryOperator,
        )

        space = self._domain
        wavenumber, npade, theta, damped_wavenumber = self.descriptor.options

        mass = identity(
            space,
            space,
            space,
            self._parameters,
            self._device_interface,
            self.descriptor.precision,
        ).weak_form()
        stiff = laplace_beltrami(
            space,
            space,
            space,
            self._parameters,
            self._device_interface,
            self.descriptor.precision,
        ).weak_form()

        if damped_wavenumber is None:
            bbox = space.grid.bounding_box
            rad = _np.linalg.norm(bbox[:, 1] - bbox[:, 0]) / 2.0
            dk = wavenumber + 0.4j * wavenumber ** (1.0 / 3.0) * rad ** (-2.0 / 3.0)
        else:
            dk = damped_wavenumber

        c0, alpha, beta, _ = _common.pade_coeffs(npade, theta)

        series = c0 * mass
        for i in range(npade):
            element = (
                alpha[i] / (dk**2) * stiff * InverseSparseDiscreteBoundaryOperator(mass - beta[i] / (dk**2) * stiff)
            )
            series -= element * mass
        operator = 1.0j * wavenumber * series

        return operator


def osrc_ntd(
    space,
    wavenumber,
    npade=2,
    theta=_np.pi / 3.0,
    damped_wavenumber=None,
    parameters=None,
    device_interface=None,
    precision=None,
):
    """Assemble the OSRC approximation to the NtD operator."""
    if space.shapeset.identifier != "p1_discontinuous":
        raise ValueError("Space shapeset must be of type 'p1_discontinuous'.")

    return _OsrcNtD(
        space,
        parameters,
        [wavenumber, npade, theta, damped_wavenumber],
        device_interface,
        precision,
    )


class _OsrcNtD(_BoundaryOperator):
    """Implementation of the OSRC NtD operator."""

    def __init__(self, space, parameters, operator_options, device_interface=None, precision=None):
        from bempp_cl.api.operators import OperatorDescriptor

        super().__init__(space, space, space, parameters)

        self._device_interface = device_interface

        self._operator_descriptor = OperatorDescriptor(
            "osrc_ntd",
            operator_options,
            "laplace_beltrami",
            "default_sparse",
            precision,
            True,
            None,
            1,
        )

    @property
    def descriptor(self):
        """Operator descriptor."""
        return self._operator_descriptor

    def _assemble(self):
        from bempp_cl.api.operators.boundary.sparse import identity
        from bempp_cl.api.operators.boundary.sparse import laplace_beltrami
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            InverseSparseDiscreteBoundaryOperator,
        )

        space = self._domain
        wavenumber, npade, theta, damped_wavenumber = self.descriptor.options

        mass = identity(
            space,
            space,
            space,
            self._parameters,
            self._device_interface,
            self.descriptor.precision,
        ).weak_form()
        stiff = laplace_beltrami(
            space,
            space,
            space,
            self._parameters,
            self._device_interface,
            self.descriptor.precision,
        ).weak_form()

        if damped_wavenumber is None:
            bbox = space.grid.bounding_box
            rad = _np.linalg.norm(bbox[:, 1] - bbox[:, 0]) / 2.0
            dk = wavenumber + 0.4j * wavenumber ** (1.0 / 3.0) * rad ** (-2.0 / 3.0)
        else:
            dk = damped_wavenumber

        c0, alpha, beta, _ = _common.pade_coeffs(npade, theta)

        series = c0 * mass
        for i in range(npade):
            element = (
                alpha[i] / (dk**2) * stiff * InverseSparseDiscreteBoundaryOperator(mass - beta[i] / (dk**2) * stiff)
            )
            series -= element * mass
        operator = (
            1.0
            / (1.0j * wavenumber)
            * (mass * InverseSparseDiscreteBoundaryOperator(mass - 1.0 / (dk**2) * stiff) * series)
        )

        return operator



"""
Actual implementation of OpenCL assemblers.
"""

import numpy as _np
import pyopencl as _cl

WORKGROUP_SIZE_GALERKIN = 16
WORKGROUP_SIZE_POTENTIAL = 128


def singular_assembler(
    device_interface,
    operator_descriptor,
    grid,
    domain,
    dual_to_range,
    test_points,
    trial_points,
    quad_weights,
    test_elements,
    trial_elements,
    test_offsets,
    trial_offsets,
    weights_offsets,
    number_of_quad_points,
    kernel_options,
    result,
):
    """Assemble singular part of integral operators with OpenCL."""
    from bempp_cl.api.utils.helpers import get_type
    from bempp_cl.core.opencl_kernels import get_kernel_from_operator_descriptor
    from bempp_cl.core.opencl_kernels import default_context, default_device

    mf = _cl.mem_flags
    ctx = default_context()
    device = default_device()

    precision = operator_descriptor.precision
    dtype = get_type(precision).real

    options = {
        "WORKGROUP_SIZE": WORKGROUP_SIZE_GALERKIN,
        "TEST": dual_to_range.shapeset.identifier,
        "TRIAL": domain.shapeset.identifier,
        "NUMBER_OF_TEST_SHAPE_FUNCTIONS": dual_to_range.number_of_shape_functions,
        "NUMBER_OF_TRIAL_SHAPE_FUNCTIONS": domain.number_of_shape_functions,
    }

    if operator_descriptor.is_complex:
        options["COMPLEX_KERNEL"] = None

    kernel = get_kernel_from_operator_descriptor(operator_descriptor, options, "singular")

    # Initialize OpenCL Buffers

    grid_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=grid.as_array.astype(dtype))
    test_normals_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=dual_to_range.normal_multipliers)
    trial_normals_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=domain.normal_multipliers)
    test_points_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=test_points.astype(dtype))
    trial_points_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=trial_points.astype(dtype))
    quad_weights_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=quad_weights.astype(dtype))
    test_elements_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=test_elements)
    trial_elements_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=trial_elements)
    test_offsets_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=test_offsets)
    trial_offsets_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=trial_offsets)
    weights_offsets_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=weights_offsets)

    local_quad_points = number_of_quad_points // WORKGROUP_SIZE_GALERKIN

    local_quad_points_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=local_quad_points)

    result_buffer = _cl.Buffer(ctx, mf.WRITE_ONLY, size=result.nbytes)

    if not kernel_options:
        kernel_options = [0.0]

    kernel_options_array = _np.array(kernel_options, dtype=dtype)

    kernel_options_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=kernel_options_array)

    number_of_singular_indices = len(test_elements)

    with _cl.CommandQueue(ctx, device=device) as queue:
        kernel(
            queue,
            (number_of_singular_indices,),
            (WORKGROUP_SIZE_GALERKIN,),
            grid_buffer,
            test_normals_buffer,
            trial_normals_buffer,
            test_points_buffer,
            trial_points_buffer,
            quad_weights_buffer,
            test_elements_buffer,
            trial_elements_buffer,
            test_offsets_buffer,
            trial_offsets_buffer,
            weights_offsets_buffer,
            local_quad_points_buffer,
            result_buffer,
            kernel_options_buffer,
            g_times_l=True,
        )
        _cl.enqueue_copy(queue, result, result_buffer)


def dense_assembler(device_interface, operator_descriptor, domain, dual_to_range, parameters, result):
    """Assemble dense with OpenCL."""
    import bempp_cl.api
    from bempp_cl.api.integration.triangle_gauss import rule
    from bempp_cl.api.utils.helpers import get_type
    from bempp_cl.core.opencl_kernels import get_kernel_from_operator_descriptor
    from bempp_cl.core.opencl_kernels import (
        default_context,
        default_device,
        get_vector_width,
    )

    if bempp_cl.api.BOUNDARY_OPERATOR_DEVICE_TYPE == "gpu":
        device_type = "gpu"
    elif bempp_cl.api.BOUNDARY_OPERATOR_DEVICE_TYPE == "cpu":
        device_type = "cpu"
    else:
        raise RuntimeError(f"Unknown device type {bempp_cl.api.BOUNDARY_OPERATOR_DEVICE_TYPE}")

    mf = _cl.mem_flags
    ctx = default_context(device_type)
    device = default_device(device_type)

    precision = operator_descriptor.precision
    dtype = get_type(precision).real
    kernel_options = operator_descriptor.options

    quad_points, quad_weights = rule(parameters.quadrature.regular)

    test_indices, test_color_indexptr = dual_to_range.get_elements_by_color()
    trial_indices, trial_color_indexptr = domain.get_elements_by_color()

    number_of_test_colors = len(test_color_indexptr) - 1
    number_of_trial_colors = len(trial_color_indexptr) - 1

    options = {
        "NUMBER_OF_QUAD_POINTS": len(quad_weights),
        "TEST": dual_to_range.shapeset.identifier,
        "TRIAL": domain.shapeset.identifier,
        "TRIAL_NUMBER_OF_ELEMENTS": domain.number_of_support_elements,
        "TEST_NUMBER_OF_ELEMENTS": dual_to_range.number_of_support_elements,
        "NUMBER_OF_TEST_SHAPE_FUNCTIONS": dual_to_range.number_of_shape_functions,
        "NUMBER_OF_TRIAL_SHAPE_FUNCTIONS": domain.number_of_shape_functions,
    }

    if operator_descriptor.is_complex:
        options["COMPLEX_KERNEL"] = None

    main_kernel = get_kernel_from_operator_descriptor(operator_descriptor, options, "regular", device_type=device_type)
    remainder_kernel = get_kernel_from_operator_descriptor(
        operator_descriptor,
        options,
        "regular",
        force_novec=True,
        device_type=device_type,
    )

    test_indices_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=test_indices)
    trial_indices_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=trial_indices)

    test_normals_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=dual_to_range.normal_multipliers)
    trial_normals_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=domain.normal_multipliers)
    test_grid_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=dual_to_range.grid.as_array.astype(dtype),
    )
    trial_grid_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=domain.grid.as_array.astype(dtype))

    test_elements_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=dual_to_range.grid.elements.ravel(order="F"),
    )

    trial_elements_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=domain.grid.elements.ravel(order="F"),
    )

    test_local2global_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=dual_to_range.local2global)

    trial_local2global_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=domain.local2global)

    test_multipliers_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=dual_to_range.local_multipliers.astype(dtype),
    )

    trial_multipliers_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=domain.local_multipliers.astype(dtype),
    )

    quad_points_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=quad_points.ravel(order="F").astype(dtype),
    )

    quad_weights_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=quad_weights.astype(dtype))

    result_buffer = _cl.Buffer(ctx, mf.READ_WRITE, size=result.nbytes)

    if not kernel_options:
        kernel_options = [0.0]

    kernel_options_array = _np.array(kernel_options, dtype=dtype)

    kernel_options_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=kernel_options_array)

    vector_width = get_vector_width(precision, device_type=device_type)

    def kernel_runner(
        queue,
        test_offset,
        trial_offset,
        test_number_of_indices,
        trial_number_of_indices,
    ):
        """Actually run the kernel for a given range."""
        remainder_size = trial_number_of_indices % vector_width
        main_size = trial_number_of_indices - remainder_size

        buffers = [
            test_indices_buffer,
            trial_indices_buffer,
            test_normals_buffer,
            trial_normals_buffer,
            test_grid_buffer,
            trial_grid_buffer,
            test_elements_buffer,
            trial_elements_buffer,
            test_local2global_buffer,
            trial_local2global_buffer,
            test_multipliers_buffer,
            trial_multipliers_buffer,
            quad_points_buffer,
            quad_weights_buffer,
            result_buffer,
            kernel_options_buffer,
            _np.int32(dual_to_range.global_dof_count),
            _np.int32(domain.global_dof_count),
            _np.uint8(domain.grid != dual_to_range.grid),
        ]

        if main_size > 0:
            main_kernel(
                queue,
                (test_number_of_indices, main_size // vector_width),
                (1, 1),
                *buffers,
                global_offset=(test_offset, trial_offset),
            )

        if remainder_size > 0:
            remainder_kernel(
                queue,
                (test_number_of_indices, remainder_size),
                (1, 1),
                *buffers,
                global_offset=(test_offset, trial_offset + main_size),
            )

    with _cl.CommandQueue(ctx, device=device) as queue:
        _cl.enqueue_fill_buffer(queue, result_buffer, _np.uint8(0), 0, result.nbytes)
        for test_index in range(number_of_test_colors):
            test_offset = test_color_indexptr[test_index]
            n_test_indices = test_color_indexptr[1 + test_index] - test_color_indexptr[test_index]
            for trial_index in range(number_of_trial_colors):
                n_trial_indices = trial_color_indexptr[1 + trial_index] - trial_color_indexptr[trial_index]
                trial_offset = trial_color_indexptr[trial_index]
                kernel_runner(queue, test_offset, trial_offset, n_test_indices, n_trial_indices)
        _cl.enqueue_copy(queue, result, result_buffer)


def potential_assembler(device_interface, space, operator_descriptor, points, parameters):
    """Assemble dense with OpenCL."""
    import bempp_cl.api
    from bempp_cl.api.integration.triangle_gauss import rule
    from bempp_cl.api.utils.helpers import get_type
    from bempp_cl.core.opencl_kernels import get_kernel_from_name
    from bempp_cl.core.opencl_kernels import get_kernel_from_operator_descriptor
    from bempp_cl.core.opencl_kernels import (
        default_context,
        default_device,
        get_vector_width,
    )

    if bempp_cl.api.POTENTIAL_OPERATOR_DEVICE_TYPE == "gpu":
        device_type = "gpu"
    elif bempp_cl.api.POTENTIAL_OPERATOR_DEVICE_TYPE == "cpu":
        device_type = "cpu"
    else:
        raise RuntimeError(f"Unknown device type {bempp_cl.api.POTENTIAL_OPERATOR_DEVICE_TYPE}")

    mf = _cl.mem_flags
    ctx = default_context(device_type)
    device = default_device(device_type)

    quad_points, quad_weights = rule(parameters.quadrature.regular)

    precision = operator_descriptor.precision
    dtype = get_type(precision).real
    kernel_options = operator_descriptor.options
    kernel_dimension = operator_descriptor.kernel_dimension

    if operator_descriptor.is_complex:
        result_type = _np.dtype(get_type(precision).complex)
    else:
        result_type = dtype
    result_type = _np.dtype(result_type)

    indices = space.support_elements
    nelements = len(indices)
    vector_width = get_vector_width(precision, device_type=device_type)
    npoints = points.shape[1]
    remainder_size = nelements % WORKGROUP_SIZE_POTENTIAL
    main_size = nelements - remainder_size

    main_kernel = None
    remainder_kernel = None
    sum_kernel = None

    options = {
        "NUMBER_OF_QUAD_POINTS": len(quad_weights),
        "SHAPESET": space.shapeset.identifier,
        "NUMBER_OF_SHAPE_FUNCTIONS": space.number_of_shape_functions,
        "WORKGROUP_SIZE": WORKGROUP_SIZE_POTENTIAL // vector_width,
    }

    if operator_descriptor.is_complex:
        options["COMPLEX_KERNEL"] = None
        options["COMPLEX_COEFFICIENTS"] = None
        options["COMPLEX_RESULT"] = None

    if main_size > 0:
        main_kernel = get_kernel_from_operator_descriptor(
            operator_descriptor, options, "potential", device_type=device_type
        )
        sum_kernel = get_kernel_from_name("sum_for_potential_novec", options, precision, device_type=device_type)

    if remainder_size > 0:
        options["WORKGROUP_SIZE"] = remainder_size
        remainder_kernel = get_kernel_from_operator_descriptor(
            operator_descriptor,
            options,
            "potential",
            force_novec=True,
            device_type=device_type,
        )

    indices_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=indices)

    normals_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=space.normal_multipliers)

    points_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=points.ravel(order="F").astype(dtype),
    )

    grid_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=space.grid.as_array.astype(dtype))

    # elements_buffer = _cl.Buffer(
    #     ctx,
    #     mf.READ_ONLY | mf.COPY_HOST_PTR,
    #     hostbuf=space.grid.elements.ravel(order="F"),
    # )

    quad_points_buffer = _cl.Buffer(
        ctx,
        mf.READ_ONLY | mf.COPY_HOST_PTR,
        hostbuf=quad_points.ravel(order="F").astype(dtype),
    )

    quad_weights_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=quad_weights.astype(dtype))

    result_buffer = _cl.Buffer(ctx, mf.READ_WRITE, size=result_type.itemsize * kernel_dimension * npoints)

    coefficients_buffer = _cl.Buffer(ctx, mf.READ_ONLY, size=result_type.itemsize * space.map_to_full_grid.shape[0])

    if main_size > 0:
        sum_size = kernel_dimension * npoints * (nelements // WORKGROUP_SIZE_POTENTIAL) * result_type.itemsize
        sum_buffer = _cl.Buffer(ctx, mf.READ_WRITE, size=sum_size)

    if not kernel_options:
        kernel_options = [0.0]

    kernel_options_array = _np.array(kernel_options, dtype=dtype)

    kernel_options_buffer = _cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=kernel_options_array)

    def evaluator(x):
        """Evaluate a potential."""
        result = _np.empty(kernel_dimension * npoints, dtype=result_type)

        with _cl.CommandQueue(ctx, device=device) as queue:
            _cl.enqueue_copy(queue, coefficients_buffer, x.astype(result_type))
            _cl.enqueue_fill_buffer(
                queue,
                result_buffer,
                _np.uint8(0),
                0,
                kernel_dimension * npoints * result_type.itemsize,
            )

            if main_size > 0:
                _cl.enqueue_fill_buffer(queue, sum_buffer, _np.uint8(0), 0, sum_size)
                queue.finish()

                main_kernel(
                    queue,
                    (npoints, main_size // vector_width),
                    (1, WORKGROUP_SIZE_POTENTIAL // vector_width),
                    grid_buffer,
                    indices_buffer,
                    normals_buffer,
                    points_buffer,
                    coefficients_buffer,
                    quad_points_buffer,
                    quad_weights_buffer,
                    sum_buffer,
                    kernel_options_buffer,
                )

                sum_kernel(
                    queue,
                    (kernel_dimension * npoints,),
                    (1,),
                    sum_buffer,
                    result_buffer,
                    _np.uint32(nelements // WORKGROUP_SIZE_POTENTIAL),
                )

            if remainder_size > 0:
                remainder_kernel(
                    queue,
                    (npoints, remainder_size),
                    (1, remainder_size),
                    grid_buffer,
                    indices_buffer,
                    normals_buffer,
                    points_buffer,
                    coefficients_buffer,
                    quad_points_buffer,
                    quad_weights_buffer,
                    result_buffer,
                    kernel_options_buffer,
                    global_offset=(0, main_size),
                )

            _cl.enqueue_copy(queue, result, result_buffer)

        return result

    return evaluator



"""
Save and Load projects
"""
import numpy as np
import os
from shutil import copy2
from os.path import join
from numpy import asanyarray as array
from electroacPy import loudspeakerSystem
from electroacPy.acousticSim.bem import bem
from electroacPy.acousticSim.pointSource import pointSource, pointSourceBEM
from electroacPy.acousticSim.evaluations import evaluations as evs
import bempp_cl.api

def save(projectPath, loudspeakerSystem):
    """
    Save loudspeaker system simulation in project folder.

    :param loudspeakerSystem:
    :return:
    *.npz archive
    """

    if not os.path.exists(projectPath):
        os.mkdir(projectPath)

    sim = loudspeakerSystem

    # global variables
    frequency = sim.frequency

    # Save lumped element / analytic simulations
    np.savez(join(projectPath, 'LEM'),
             frequency   = frequency,
             driver      = sim.driver,
             laser       = sim.vibrometry,
             crossover   = sim.crossover,
             enclosure   = sim.enclosure,
             radiator_id = sim.radiator_id,
             c           = sim.c,
             rho         = sim.rho)

    # Save evaluations / bem
    for study in sim.acoustic_study:
        if bool(sim.evaluation) is True:
            obsTmp = sim.evaluation[study]
            np.savez(join(projectPath, 'evs_{}'.format(study)),
                     frequency            = obsTmp.frequency,
                     setup                = obsTmp.setup,
                     referenceStudy       = obsTmp.referenceStudy,)
        else:
            pass

        # BEM studies
        # studyTmp = sim.acoustic_study[study]
        # pressureArrayAcs = storePressureMeshResults(studyTmp)
        # copy2(studyTmp.meshPath, projectPath)
        # mesh_filename = os.path.basename(studyTmp.meshPath)
        
        studyTmp = sim.acoustic_study[study]
        
        # prepare to store mesh pressure and velocity 
        if studyTmp.admittanceCoeff is not None and hasattr(studyTmp, "xSource"):
            pressureArrayAcs, velocityArrayAcs, velocityArrayAcs_Y = storePressureMeshResults_PSINT(studyTmp)
        else:
            pressureArrayAcs, velocityArrayAcs = storePressureMeshResults(studyTmp)
            velocityArrayAcs_Y = None
            
        # do a copy of mesh in save folder
        if studyTmp.meshPath is not None:
            copy2(studyTmp.meshPath, projectPath)
            mesh_filename = os.path.basename(studyTmp.meshPath)
        else:
            mesh_filename = None
            
        # EXTERIOR
        if hasattr(studyTmp, "xSource"):
            np.savez(join(projectPath, 'acs_{}'.format(study)),
                     meshPath           = studyTmp.meshPath,
                     mesh_filename      = mesh_filename,
                     radiatingElement   = studyTmp.radiatingElement,
                     QSource            = studyTmp.QSource,
                     xSource            = studyTmp.xSource,
                     isComputed         = studyTmp.isComputed,
                     vertices           = studyTmp.vertices,
                     domain             = studyTmp.domain,
                     kwargs             = studyTmp.kwargs,
                     pressureArrayAcs   = pressureArrayAcs,
                     velocityArrayAcs   = velocityArrayAcs,
                     velocityArrayAcs_Y = velocityArrayAcs_Y,
                     LEM_enclosures     = studyTmp.LEM_enclosures,
                     radiator           = studyTmp.radiator,
                     c_0                = studyTmp.c_0,
                     rho_0              = studyTmp.rho_0
                     )
        else:
            np.savez(join(projectPath, 'acs_{}'.format(study)),
                     meshPath             = studyTmp.meshPath,
                     mesh_filename        = mesh_filename,
                     radiatingElement     = studyTmp.radiatingElement,
                     velocity             = array(studyTmp.velocity, dtype=object),
                     isComputed           = studyTmp.isComputed,
                     coeff_radSurf        = studyTmp.coeff_radSurf,
                     vertices             = studyTmp.vertices,
                     domain               = studyTmp.domain,
                     kwargs               = studyTmp.kwargs,
                     pressureArrayAcs     = pressureArrayAcs,
                     LEM_enclosures       = studyTmp.LEM_enclosures,
                     radiator             = studyTmp.radiator,
                     c_0                  = studyTmp.c_0,
                     rho_0                = studyTmp.rho_0
                     )
    return None

def load(pathToProject):
    """
    Load loudspeaker simulation project from directory path.

    Parameters
    ----------
    pathToProject: str,
        path to directory where npz files are stored

    Returns
    -------
    LS: loudspeakerSystem object,
        Study, evaluations and LEM setup.
    """

    dataLEM = np.load(join(pathToProject, 'LEM.npz'), allow_pickle=True)

    # create loudspeaker system
    LS             = loudspeakerSystem(dataLEM['frequency'])
    LS.driver      = dataLEM['driver'].item()
    LS.vibrometry  = dataLEM['laser'].item()
    LS.enclosure   = dataLEM['enclosure'].item()
    LS.crossover   = dataLEM['crossover'].item()
    LS.radiator_id = dataLEM['radiator_id'].item()
    
    try: # to keep it compatible with previous datasets
        LS.c   = dataLEM['c']
        LS.rho = dataLEM['rho']
    except:
        LS.c   = 343
        LS.rho = 1.22

    # import studies and evaluations
    file_list  = os.listdir(pathToProject)
    acs_files  = [file for file in file_list if file.startswith('acs')]
    study_name = []
    for i in range(len(acs_files)):
        study_name.append(acs_files[i][4:-4])

    for study in study_name:
        # load acoustic_study
        data_acs     = np.load(join(pathToProject, 'acs_{}.npz'.format(study)),
                               allow_pickle=True)
        meshName     = data_acs['mesh_filename'].item()
        isComputed   = data_acs['isComputed'].item()
        domain       = data_acs["domain"].item()
        kwargs       = data_acs['kwargs'].item()

        try:
            enclosures   = list(data_acs['LEM_enclosures'])
        except:
            enclosures    = {}
        try:
            radiator   = list(data_acs['radiator'])
        except:
            radiator    = {}
    
        if "xSource" in data_acs and meshName is not None: # pointSourceBEM
            xSource = data_acs["xSource"]
            QSource = data_acs["QSource"]
            physics_acs = pointSourceBEM(join(pathToProject, meshName),
                                         xSource, QSource, 
                                         LS.frequency, domain=domain, 
                                         c_0=LS.c, rho_0=LS.rho, 
                                         **kwargs)
            physics_acs.isComputed     = isComputed
            physics_acs.LEM_enclosures = enclosures
            physics_acs.radiator       = radiator
            loadPointSourceBEM(physics_acs, data_acs["pressureArrayAcs"],
                               data_acs["velocityArrayAcs"], 
                               data_acs["velocityArrayAcs_Y"])
            LS.acoustic_study[study] = physics_acs
            
        elif "xSource" in data_acs and meshName is None: # pointSource
            xSource = data_acs["xSource"]
            QSource = data_acs["QSource"]
            physics_acs = pointSource(xSource, QSource, LS.frequency,
                                      c=LS.c, rho=LS.rho,
                                      boundary_conditions=kwargs["boundary_conditions"],
                                      **kwargs)
            physics_acs.isComputed = isComputed
            physics_acs.LEM_enclosures = enclosures
            physics_acs.radiator       = radiator
            LS.acoustic_study[study] = physics_acs
        else: # BEM
            radSurf      = data_acs['radiatingElement']
            surfVelocity = data_acs['velocity']  
            physics_acs = bem(join(pathToProject, meshName),
                                  radSurf,
                                  surfVelocity,
                                  LS.frequency,
                                  domain=domain,
                                  c_0=LS.c, rho_0=LS.rho,
                                  **kwargs)
            physics_acs.isComputed     = isComputed
            physics_acs.LEM_enclosures = enclosures
            physics_acs.radiator       = radiator
            loadPressureMeshResults(physics_acs, data_acs['pressureArrayAcs'])
            LS.acoustic_study[study] = physics_acs

        # load evaluations
        try:
            try:
                data_evs = np.load(join(pathToProject, 'evs_{}.npz'.format(study)), allow_pickle=True)
            except:
                data_evs = np.load(join(pathToProject, 'obs_{}.npz'.format(study)), allow_pickle=True)
                print("Legacy evs loaded")
            physics_evs = evs(physics_acs)
            physics_evs.setup = data_evs["setup"].item()
            physics_evs.frequency = data_evs["frequency"]
            physics_evs.referenceStudy = data_evs["referenceStudy"].item()
            LS.evaluation[study] = physics_evs
        except:
            print('No evaluation to load')
    return LS


#%% HELPERS
def storePressureMeshResults_PSINT(acoustic_study):
    study = acoustic_study
    Nfft = len(study.frequency)
    nRad = study.Ns
    nCoeff = len(study.p_mesh[0, 0].coefficients)
    nCoeffV = len(study.u_mesh[0, 0].coefficients)
    
    # store pressure
    pressureMesh = np.zeros([Nfft, nRad, nCoeff], dtype=complex)
    velocityMesh = np.zeros([Nfft, nRad, nCoeffV], dtype=complex)
    velocityMesh_Y = np.zeros([Nfft, nRad, nCoeffV], dtype=complex)
    for freq in range(Nfft):
        for rad in range(nRad):
            pressureMesh[freq, rad, :] = study.p_mesh[freq, rad].coefficients
            velocityMesh[freq, rad, :] = study.u_mesh[freq, rad].coefficients
            velocityMesh_Y[freq, rad, :] = study.u_mesh_Y[freq, rad].coefficients
    return pressureMesh, velocityMesh, velocityMesh_Y

def storePressureMeshResults(acoustic_study):
    study = acoustic_study
    Nfft = len(study.frequency)
    nRad = study.Ns
    nCoeff = len(study.p_mesh[0, 0].coefficients)
    nCoeffV = len(study.u_mesh[0, 0].coefficients)
    
    # store pressure
    pressureMesh = np.zeros([Nfft, nRad, nCoeff], dtype=complex)
    velocityMesh = np.zeros([Nfft, nRad, nCoeffV], dtype=complex)
    for freq in range(Nfft):
        for rad in range(nRad):
            pressureMesh[freq, rad, :] = study.p_mesh[freq, rad].coefficients
            # velocityMesh[freq, rad, :] = study.u_mesh[freq, rad].coefficients
    return pressureMesh, velocityMesh


# LOAD PRESSURE
def loadPressureMeshResults(obj, pressureMesh):
    Nfft  = np.shape(pressureMesh)[0]
    nRad  = np.shape(pressureMesh)[1]

    for f in range(Nfft):
        for rs in range(nRad):
            obj.p_mesh[f, rs] = bempp_cl.api.GridFunction(obj.spaceP, coefficients=pressureMesh[f, rs, :])
            dofCount = obj.spaceU_freq[rs].grid_dof_count
            coeff_radSurf = np.ones(dofCount, dtype=complex) * obj.coeff_radSurf[f, rs, :dofCount]
            spaceU = bempp_cl.api.function_space(obj.grid_sim, "DP", 0,
                                              segments=[obj.radiatingElement[rs]])
            u_total = bempp_cl.api.GridFunction(spaceU, coefficients=-coeff_radSurf)
            obj.u_mesh[f, rs] = u_total

        obj.p_total_mesh[f] = bempp_cl.api.GridFunction(obj.spaceP,
                                                     coefficients=np.sum(pressureMesh[f, :, :], 0))
    return None


def loadPointSourceBEM(obj, pressureMesh, velocityMesh, velocityMesh_Y):
    Nfft  = np.shape(pressureMesh)[0]
    nRad  = np.shape(pressureMesh)[1]
    for f in range(Nfft):
        for rs in range(nRad):
            obj.p_mesh[f, rs] = bempp_cl.api.GridFunction(obj.spaceP, 
                                                       coefficients=pressureMesh[f, rs, :])
            obj.u_mesh[f, rs] = bempp_cl.api.GridFunction(obj.spaceP, 
                                                       coefficients=velocityMesh[f, rs, :])        
            if obj.admittanceCoeff is not None:
                obj.u_mesh_Y[f, rs] = bempp_cl.api.GridFunction(obj.spaceP, 
                                                             coefficients=velocityMesh_Y[f, rs, :])
    return None



"""Interface to DOLFINX for FEM-BEM coupling."""


def boundary_grid_from_fenics_mesh(fenics_mesh):
    """
    Create a Bempp boundary grid from a FEniCS Mesh.

    Return the Bempp grid and a map from the node numberings of the FEniCS
    mesh to the node numbers of the boundary grid.
    """
    import bempp_cl.api
    import numpy as np
    from dolfinx.mesh import entities_to_geometry, exterior_facet_indices

    fenics_mesh.topology.create_entities(2)
    fenics_mesh.topology.create_connectivity(2, 3)
    fenics_mesh.topology.create_entity_permutations()

    facets = exterior_facet_indices(fenics_mesh.topology)
    boundary = entities_to_geometry(
        fenics_mesh,
        fenics_mesh.topology.dim - 1,
        facets,
        True,
    )

    c23 = fenics_mesh.topology.connectivity(2, 3)

    tet_indices = np.array([c23.links(facet)[0] for facet in facets])
    tet_vertices = entities_to_geometry(
        fenics_mesh,
        fenics_mesh.topology.dim,
        tet_indices,
        True,
    )

    bm_nodes = list(set(node for tri in boundary for node in tri))
    bm_cells = np.zeros([3, len(boundary)])
    for i, (tri, tet) in enumerate(zip(boundary, tet_vertices)):
        for j in range(3):
            bm_cells[j, i] = bm_nodes.index(tri[j])
        v0 = fenics_mesh.geometry.x[tri[0]]
        v1 = fenics_mesh.geometry.x[tri[1]]
        v2 = fenics_mesh.geometry.x[tri[2]]
        v3 = fenics_mesh.geometry.x[[i for i in tet if i not in tri][0]]
        to_other_vertex = v3 - v0
        normal = np.cross(v1 - v0, v2 - v0)
        if np.dot(normal, to_other_vertex) > 0:
            bm_cells[1, i], bm_cells[2, i] = bm_cells[2, i], bm_cells[1, i]

    bm_coords = fenics_mesh.geometry.x[bm_nodes].transpose()

    bempp_boundary_grid = bempp_cl.api.Grid(bm_coords, bm_cells)

    return bempp_boundary_grid, bm_nodes


def fenics_to_bempp_trace_data(fenics_space):
    """Return tuple (space,trace_matrix)."""
    family, degree = fenics_space_info(fenics_space)

    if family in ["Lagrange", "P"]:
        if degree == 1:
            return p1_trace(fenics_space)
    else:
        raise NotImplementedError()


def fenics_space_info(fenics_space):
    """Return tuple (family,degree) containing information about a FEniCS space."""
    element = fenics_space.ufl_element()
    family = element.basix_element.family.name
    degree = element.degree
    return family, degree


# pylint: disable=too-many-locals
def p1_trace(fenics_space):
    """
    Return the P1 trace operator.

    This function returns a pair (space, trace_matrix),
    where space is a Bempp space object and trace_matrix is the corresponding
    matrix that maps the coefficients of a FEniCS function to its boundary
    trace coefficients in the corresponding Bempp space.
    """
    import bempp_cl.api
    from scipy.sparse import coo_matrix
    import numpy as np

    family, degree = fenics_space_info(fenics_space)
    if not (family in ["Lagrange", "P"] and degree == 1):
        raise ValueError("fenics_space must be a p1 Lagrange space")

    fenics_mesh = fenics_space.mesh
    bempp_boundary_grid, bm_nodes = boundary_grid_from_fenics_mesh(fenics_mesh)

    # First get trace space
    space = bempp_cl.api.function_space(bempp_boundary_grid, "P", 1)

    num_fenics_vertices = fenics_mesh.topology.connectivity(0, 0).num_nodes

    # FEniCS vertices to bempp dofs
    b_vertices_from_vertices = coo_matrix(
        (np.ones(len(bm_nodes)), (np.arange(len(bm_nodes)), bm_nodes)),
        shape=(len(bm_nodes), num_fenics_vertices),
        dtype="float64",
    ).tocsc()

    # Finally FEniCS dofs to vertices.
    dof_to_vertex_map = np.zeros(num_fenics_vertices, dtype=np.int64)
    tets = fenics_mesh.geometry.dofmap

    # Support older versions of FEniCSx
    if hasattr(tets, "get_links"):
        ntets = tets.num_nodes
        tets = [tets.get_links(i) for i in range(ntets)]
    if hasattr(tets, "links"):
        ntets = tets.num_nodes
        tets = [tets.links(i) for i in range(ntets)]

    for tet, cell_verts in enumerate(tets):
        cell_dofs = fenics_space.dofmap.cell_dofs(tet)
        for v in range(4):
            vertex_n = cell_verts[v]
            dof = cell_dofs[fenics_space.dofmap.dof_layout.entity_dofs(0, v)[0]]
            dof_to_vertex_map[dof] = vertex_n

    vertices_from_fenics_dofs = coo_matrix(
        (
            np.ones(num_fenics_vertices),
            (dof_to_vertex_map, np.arange(num_fenics_vertices)),
        ),
        shape=(num_fenics_vertices, num_fenics_vertices),
        dtype="float64",
    ).tocsc()

    # Get trace matrix by multiplication
    trace_matrix = b_vertices_from_vertices @ vertices_from_fenics_dofs

    # Now return everything
    return space, trace_matrix


class FenicsOperator(object):
    """Wrap a FEniCS-X Operator into a Bempp operator."""

    def __init__(self, fenics_weak_form):
        """Construct an operator from a weak form in FEniCS."""
        self._fenics_weak_form = fenics_weak_form
        self._sparse_mat = None

    def weak_form(self):
        """Return the weak form."""
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            SparseDiscreteBoundaryOperator,
        )
        from dolfinx.fem import assemble_matrix, form
        from scipy.sparse import csr_matrix

        if self._sparse_mat is None:
            mat = assemble_matrix(form(self._fenics_weak_form))
            try:
                mat.scatter_reverse()
            except AttributeError:
                # Support for older FEniCSx
                mat.finalize()
            shape = tuple(
                i._ufl_function_space.dofmap.index_map.size_global for i in self._fenics_weak_form.arguments()
            )
            self._sparse_mat = csr_matrix((mat.data, mat.indices, mat.indptr), shape=shape)

        return SparseDiscreteBoundaryOperator(self._sparse_mat)



"""
Collection of acoustic circuit blocks 

"""

from electroacPy.circuitSolver.components.electric import resistance, capacitance, inductance
from electroacPy.circuitSolver.components.acoustic import cavity, port, open_line_T
from numpy import sqrt, pi
import random, string



def randomblock_id(length):
   letters = string.ascii_lowercase
   return ''.join(random.choice(letters) for i in range(length))

class sealedEnclosure:
    def __init__(self, A, Vb, fs, Vas, Cas,
                 Ql=10, Qa=100, rho=1.22, c=343):
        """
        Create a sealed enclosure. Uses loudspeaker compliance to determine 
        losses.

        Parameters
        ----------
        A : int or str,
            Input connection.
        Vb : float,
            enclosure volume.
        fs: float, 
            resonance frequency of drive unit.
        Vas: float, 
            equivalent volume of drive unit.
        Cas: float, 
            compliance of suspension in acoustic domain.
        Ql: float,
            Q factor related to leaks. (low leaks) 5 < Ql < 30 (high leaks)
        Qa: float,
            Q factor related to damping in the enclosure. (high-damping) 5 < Qa < +100 (low damping)
            if Q = 1/Cab -> no damping at all.
        
    

        Returns
        -------
        None.kwargs

        """
                
        np = str(A)
        nm = 0
        rnd_id = randomblock_id(3)

        # parameter computation
        fb = fs * sqrt(1 + Vas / Vb)
        wb = 2*pi*fb
        
        Cab = Vb / rho / c**2
        Rab = 1 / wb / Qa  / Cab
        Ral = Ql / wb / Cab
        
        self.network = {"Rab": resistance(np, np+rnd_id, Rab),
                        "Cab": capacitance(np+rnd_id, nm, Cab),
                        "Ral": resistance(np, nm, Ral)}
        
        
class portedEnclosure:
    def __init__(self, A, B, Vb, Lp, rp,
                 Ql=10, mu=1.86e-5, rho=1.22, c=343):
        """
        Create a ported enclosure. Uses loudspeaker compliance to determine 
        losses.

        Parameters
        ----------
        A : int or str,
            Input connection.
        Vb : float,
            enclosure volume.
        Ql: float,
            Q factor related to leaks. (low leaks) 5 < Ql < 30 (high leaks)
        Qa: float,
            Q factor related to damping in the enclosure. (high-damping) 5 < Qa < +100 (low damping)
            if Q = 1/Cab -> no damping at all.
        k: float,
            length correction. (one flanged termination) 0.6 < k < 0.9 (both termination are flanged)

        Returns
        -------
        None.kwargs
        """
        
        np = str(A)
        nm = str(B)
        rnd_id = randomblock_id(3)

        Sp = pi*rp**2
        
        # parameter computation
        Cab = Vb / rho / c**2
        Lt  = Lp + 0.73*rp*2     # Lp + k * sqrt(Sp/pi)
        Mp  = Lt * rho / Sp
        
        fb = 1 / (2*pi*sqrt(Mp*Cab))
        wb = 2*pi*fb
        
        Ral = Ql / wb / Cab
        # Ral = rho * c / 1e-5 / Vb

        print("Ral = ", Ral)
        print("Mp = ", Mp)
        print("Rp: ",  (2*10*2*pi*rho*mu)/Sp * (Lt/rp + 1*0.7))
        
        self.network = {"Cab": capacitance(np, 0, Cab),
                        "Ral": resistance(np, 0, Ral),
                        "Port": port(np, nm, Lp, rp, rho, c, mu)}
                        # "Port": open_line_T(np, np+"_1_"+rnd_id, nm, 0, 
                        #                     Lt, Sp)}
        
    


"""Define interfaces to external viewers."""

import numpy as _np


def visualize(obj, mode=None, transformation=None):
    """
    Create a visualisation.

    Attributes
    ----------
    obj : Grid or GridFunction object
        A grid or grid function to plot
    mode : string
        One of 'element' or 'node'. If 'element' is chosen
        the color is determined by the mid-point of the faces
        of the grid. For 'node' the vertex values are
        chosen. Only used for grid functions.
    transformation : string or object
        One of 'real', 'imag', 'abs', 'log_abs' or
        'abs_squared' or a callable object.
        Describes the data transformation
        before plotting. For functions with vector values
        only 'abs', 'log_abs' or 'abs_squared' are allowed.
        If a callable object is given this is applied instead.
        It is important that the callable returns numpy arrays
        with the same number of dimensions as before.

    """
    import bempp_cl.api
    import numpy as np

    transform = None

    if transformation == "real":
        transform = np.real
    elif transformation == "imag":
        transform = np.imag
    elif transformation == "abs":
        transform = lambda x: np.sqrt(np.sum(np.abs(x) ** 2, axis=0, keepdims=True))
    elif transformation == "log_abs":
        transform = lambda x: np.log(np.sqrt(np.sum(np.abs(x) ** 2, axis=0, keepdims=True)))
    elif transformation == "abs_squared":
        transform = lambda x: np.sum(np.abs(x) ** 2, axis=0, keepdims=True)
    else:
        transform = transformation

    if bempp_cl.api.PLOT_BACKEND == "gmsh":
        visualize_with_gmsh(obj, mode, transform)
    if bempp_cl.api.PLOT_BACKEND == "paraview":
        visualize_with_paraview(obj, mode, transform)
    if bempp_cl.api.PLOT_BACKEND == "jupyter_notebook":
        visualize_with_jupyter_notebook(obj, mode, transform)


def visualize_with_jupyter_notebook(obj, mode=None, transformation=None):
    """View a grid or grid function in an IPython Notebook."""
    import plotly
    import plotly.figure_factory as ff
    import plotly.graph_objs as go
    from bempp_cl.api import GridFunction
    from bempp_cl.api.grid.grid import Grid
    import numpy as np

    if transformation is None:
        transformation = np.real

    plotly.offline.init_notebook_mode()

    if isinstance(obj, Grid):
        vertices = obj.vertices
        elements = obj.elements
        fig = ff.create_trisurf(
            x=vertices[0, :],
            y=vertices[1, :],
            z=vertices[2, :],
            simplices=elements.T,
            color_func=elements.shape[1] * ["rgb(255, 222, 173)"],
        )
        fig["layout"]["scene"].update(go.layout.Scene(aspectmode="data"))
        plotly.offline.iplot(fig)

    elif isinstance(obj, GridFunction):
        import matplotlib as mpl
        from matplotlib import pyplot as plt

        cmap = plt.get_cmap("jet")

        grid = obj.space.grid
        vertices = grid.vertices
        elements = grid.elements

        local_coordinates = _np.array([[1.0 / 3], [1.0 / 3]])
        values = _np.zeros(grid.entity_count(0), dtype="float64")
        for element in grid.entity_iterator(0):
            index = element.index
            local_values = np.real(transformation(obj.evaluate(index, local_coordinates)))
            values[index] = local_values.flatten()

        norm = mpl.colors.Normalize(vmin=_np.min(values), vmax=_np.max(values))
        colorfun = lambda x: _np.rint(_np.array(cmap(norm(x))) * 255)
        color_codes = ["rgb({0}, {1}, {2})".format(*colorfun(x)) for x in values]
        fig = ff.create_trisurf(
            x=vertices[0, :],
            y=vertices[1, :],
            z=vertices[2, :],
            simplices=elements.T,
            color_func=color_codes,
        )
        fig["layout"]["scene"].update(go.layout.Scene(aspectmode="data"))
        plotly.offline.iplot(fig)


def visualize_with_gmsh(obj, mode=None, transformation=None):
    """
    View a grid or grid function with Gmsh.

    Parameters
    ----------
    obj : bempp_cl.api.Grid or bempp_cl.api.GridFunction
        Grid or grid function to visualize.
    mode : string
        One of 'element' or 'node'
        (default 'vertices')
    transformation : callable
        A function object that is applied to the data before
        writing it out

    Notes
    -----
    This function writes the data into a temp file and
    visualizes it.

    """
    import tempfile
    import subprocess
    from bempp_cl.api import export, GMSH_PATH, TMP_PATH, GridFunction
    from bempp_cl.api.grid.grid import Grid

    if GMSH_PATH is None:
        print("Gmsh not available for visualization.")
        return None

    outfile = tempfile.NamedTemporaryFile(suffix=".msh", dir=TMP_PATH, delete=False)
    if isinstance(obj, Grid):
        export(outfile.name, grid=obj)
    elif isinstance(obj, GridFunction):
        export(
            outfile.name,
            grid_function=obj,
            transformation=transformation,
            data_type=mode,
        )
    outfile.close()

    subprocess.Popen([GMSH_PATH, outfile.name])


def visualize_with_paraview(obj, mode=None, transformation=None):
    """
    View a grid or grid function with Paraview.

    Parameters
    ----------
    obj : bempp_cl.api.Grid or bempp_cl.api.GridFunction
        Grid or grid function to visualize.
    mode : string
        One of 'element' or 'node'
        (default 'vertices')
    transformation : callable
        A function object that is applied to the data before
        writing it out

    Notes
    -----
    This function writes the data into a temp file and
    visualizes it.

    """
    import tempfile
    import subprocess
    from bempp_cl.api import export, TMP_PATH, GridFunction
    from bempp_cl.api.grid.grid import Grid
    from bempp_cl.api.utils import which
    import os

    if os.name == "nt":
        pview = which("paraview.exe")
    else:
        pview = which("paraview")

    if pview is None:
        raise EnvironmentError("Could not find Paraview." + "Interactive plotting with Paraview not available.")

    outfile = tempfile.NamedTemporaryFile(suffix=".vtu", dir=TMP_PATH, delete=False)
    if isinstance(obj, Grid):
        export(outfile.name, grid=obj)
    elif isinstance(obj, GridFunction):
        export(
            outfile.name,
            grid_function=obj,
            transformation=transformation,
            data_type=mode,
        )
    outfile.close()

    subprocess.Popen([pview, outfile.name])


def enable_gmsh_viewer():
    """Change plotting default to Gmsh."""
    import bempp_cl.api

    bempp_cl.api.PLOT_BACKEND = "gmsh"


def enable_paraview_viewer():
    """Change plotting default to Paraview."""
    import bempp_cl.api

    bempp_cl.api.PLOT_BACKEND = "paraview"


def enable_jupyter_viewer():
    """Change plotting default to IPython."""
    import bempp_cl.api

    bempp_cl.api.PLOT_BACKEND = "jupyter_notebook"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  3 15:39:33 2022

Toolbox for electroacoustic simulations

@author: tom.munoz
"""

# =============================================================================
# Initialisation file
# =============================================================================
#%% loudspeakerSystem modules
from electroacPy.loudspeakerSystem import loudspeakerSystem
from electroacPy.io import save, load

#%% circuitSolver modules
from electroacPy.circuitSolver.solver import circuit
from electroacPy.circuitSolver import components as csc
from electroacPy.circuitSolver import blocks as csb

#%% "general" module
from electroacPy import general as gtb




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep  3 11:26:36 2024

@author: tom
"""
from numpy import unique, zeros, squeeze, linalg, concatenate, int16
from numpy import max as nmax
from copy import copy
from tqdm import tqdm
    
class circuit:
    def __init__(self, frequency):
        """
        Create a circuit object.

        Parameters
        ----------
        frequency : numpy array
            Range of simulation.

        Returns
        -------
        None.

        """
        self.frequency = frequency
        self.Nfft = len(frequency)
        self.G_m = None
        self.I_s = None
        self.X = None
        
        # component count
        self.node_count = 1
        self.node_list = None
        self.N = 0  # number of nodes
        self.M = 0  # number of independant voltage source
        self.Nc = 0 # total number of components
        self.components = []
        
        # id component
        self.node_id = {}
        self.block = []
        self.source_id = {}
        
    def addComponent(self, *args):
        """
        Add one or multiple component object into the circuit
    
        Parameters
        ----------
        component : component object
    
        Returns
        -------
        None.
    
        """
        
        for comp in args:
            comp.init_component(self.frequency)
            self.components.append(comp)
            self.checkNodes(comp)
                
    def checkNodes(self, component):
        """
        Looks for already existing node in circuit (stored in node_id). If none
        is found, add a new key to the node_id dictionary.

        Parameters
        ----------
        component : component object
            Electric, mechanic or acoustic components.

        Returns
        -------
        None.

        """
        
        # Check 'np' key and assign node if it doesn't exist
        if component.np != '0' and component.np not in self.node_id:
            self.node_id[component.np] = self.node_count
            self.node_count += 1
    
        # Check 'nm' key and assign node if it doesn't exist
        if component.nm != '0' and component.nm not in self.node_id:
            self.node_id[component.nm] = self.node_count
            self.node_count += 1
    
        # Check 'np1' key if it exists in the component and assign node if it doesn't exist
        if hasattr(component, 'np1') and component.np1 != '0' and component.np1 not in self.node_id:
            self.node_id[component.np1] = self.node_count
            self.node_count += 1
    
        # Check 'nm1' key if it exists in the component and assign node if it doesn't exist
        if hasattr(component, 'nm1') and component.nm1 != '0' and component.nm1 not in self.node_id:
            self.node_id[component.nm1] = self.node_count
            self.node_count += 1
                
                
    def addBlock(self, *args):
        """
        Add one or multiple block object into the circuit
    
        Parameters
        ----------
        block : block object
    
        Returns
        -------
        None.
    
        """
        for block in args:
            for comp in block.network:
                block.network[comp].init_component(self.frequency)
                self.components.append(block.network[comp])
                self.checkNodes(block.network[comp])

                
    def countNodesAndSources(self):                   
        """
        Count and identify nodes and sources present in the circuits.

        Returns
        -------
        None.

        """
        self.N = len(self.node_id)        
        self.Nc = len(self.components)
        
        for comp in self.components:
            if comp.vsource != 0:
                self.M += comp.vsource
        
        nbsource = 0
        for comp in self.components:
            if "Is" not in comp.contribute:
                comp.update_stamp(self.node_id, self.M)
            elif "Is" in comp.contribute:
                comp.update_stamp(self.node_id, self.M, nbsource)
                nbsource += comp.vsource  # add number of sources
                # print("nbsource", nbsource)
                # print("comp.np", comp.np)
                if comp.vsource == 2:
                    self.source_id[comp.np] = nbsource-1
                    self.source_id[comp.np1] = nbsource
                else:
                    self.source_id[comp.np] = nbsource
    
    def build_G(self, nf):
        G = zeros([self.N+self.M, self.M+self.N], dtype=complex)
        for comp in self.components:
            if "G" in comp.contribute:
                G += comp.stamp_G * comp.Gs[nf]
                if "Ys" in comp.__dict__:
                    G += comp.stamp_Y * comp.Ys[nf]             
        return G
    
    def build_Is(self, nf):
        I = zeros([self.N + self.M, 1], dtype=complex)
        for comp in self.components:
            if "Is" in comp.contribute:
                I += comp.stamp_I * comp.Gs[nf]
        return I
    
    def run(self):
        self.countNodesAndSources()
        self.X = zeros([self.N+self.M, self.Nfft], dtype=complex)
        
        print("Solving network...")
        for nf in tqdm(range(self.Nfft)):
            self.G_m = self.build_G(nf)
            self.I_s = self.build_Is(nf)
            self.X[:, nf] = squeeze(linalg.inv(self.G_m) @ self.I_s)
    
    def getPotential(self, node_id):
        """
        Return the potential at given node.

        Parameters
        ----------
        node_id : int or str
            Node at which the potential is probed.

        Returns
        -------
        out : numpy array
            Voltage at node "node_id".

        """

        if isinstance(node_id, int):
            node_id = str(node_id)
        node_number = self.node_id[node_id]
        
        if self.X is not None:
            if node_number <= self.N:
                out = self.X[node_number-1]
            else:
                raise Exception("Potential is given for N nodes: node_number <= N")
        return out
    
    def getFlow(self, source_id):
        """
        Return the flow at given source.

        Parameters
        ----------
        source_id : int or str
            Node of source, as defined by the user when creating the component.

        Returns
        -------
        out : numpy array
            Current at source "source_id".

        """

        if isinstance(source_id, int):
            source_id = str(source_id)
        source_number = self.source_id[source_id]
        
        if self.X is not None:
            if source_number <= self.M:
                out = self.X[self.N+source_number-1]
            else:
                raise Exception("Flow is given for M sources: source_number <= M")
    
        return out
     
    


"""
Helpful plotting functions.

"""

import numpy as np
import electroacPy.general as gtb
import matplotlib.pyplot as plt
from matplotlib.widgets import Cursor
from matplotlib.backend_bases import MouseButton
import pyvista


#%% helper
def sumPressureArray(bemObj, radiatingSurface, radiationCoeff=None):
    p_mesh = bemObj.p_mesh
    radSurf_system = bemObj.radiatingElement
    if isinstance(radiatingSurface, str):
        if radiatingSurface == 'all':
            radiatingSurface = radSurf_system
    else:
        radiatingSurface = np.array(radiatingSurface)

    pressureCoeff = np.zeros([len(bemObj.frequency), 
                              len(bemObj.p_mesh[0, 0].coefficients)],
                             dtype='complex')
    
    if radiationCoeff is None:
        radiationCoeff = np.ones([len(bemObj.frequency), 
                                  len(radiatingSurface)], dtype=complex)
    for f in range(len(bemObj.frequency)):
        for j in range(len(radiatingSurface)):
            ind_surface = np.argwhere(radSurf_system == radiatingSurface[j])[0][0]
            pressureCoeff[f, :] += (p_mesh[f][ind_surface].coefficients * 
                                    radiationCoeff[f, j])
    return pressureCoeff


# %%2D plots
def FRF(freq, H, transformation="SPL", logx=True, legend=None, **kwargs):  #logx=True, xlim=None, ylim=None, title=None, save=False, dpi=64):
    """
    Plot Frequency Response Functions (FRFs).

    Parameters:
        freq (tuple or array): Frequency array or tuple of frequency arrays.
        H (tuple or array): FRF array or tuple of FRF arrays. If providing multiple FRF arrays,
                            each element should correspond to the FRF for the respective frequency array.
                            If a single FRF array is provided, it will be plotted against each frequency array.
        transformation (str, optional): Type of transformation to apply to the FRF data before plotting.
                                        Defaults to "SPL".
        labels (tuple or str, optional): Label(s) for the FRF(s) being plotted. If providing multiple FRFs,
                                         should be a tuple of strings or a list of strings, with each element
                                         corresponding to the label for the respective FRF(s). Defaults to None.

    Returns:
        None
    """
    from electroacPy.general.gain import SPL, dB

    # get kwargs
    labels = legend
    # check inputs
    if isinstance(freq, tuple):
        freq = freq
    else:
        freq = (freq,)

    if isinstance(H, tuple):
        H = H
    else:
        H = (H,)

    if isinstance(labels, tuple):
        labels = labels
    elif labels == None:
        pass
    else:
        labels = (labels,) * len(freq)

    if len(freq) == 1:
        if len(H) > 1:
            freq_out = freq
            for i in range(len(H)-1):
                freq_out += freq
            freq = freq_out

    # associate transformations
    if transformation in ["SPL", "spl"]:
        tr = SPL
        tr_str = "SPL [dB]"
    elif transformation == "dB":
        tr = dB
        tr_str = "gain [dB]"
    elif transformation == "abs":
        tr = np.abs
        tr_str = "magnitude [abs]"
    elif transformation == "real":
        tr = np.real
        tr_str = "Real"
    elif transformation == "imag":
        tr = np.imag
        tr_str = "Imaginary"
    elif transformation == "phase":
        tr = np.angle
        tr_str = "Phase [rad]"
    elif transformation == "unwrap_phase":
        tr = lambda x: np.unwrap(np.angle(x))
        tr_str = "Phase [rad]"
    else:
        raise Exception("transformation not understood, "
                        "available transformations: 'SPL', 'dB', 'abs', 'phase', 'real', 'imag'")

    plt.figure(figsize=kwargs.get('figsize', None))
    for i in range(len(freq)):
        if isinstance(H[i], list):
            for j in range(len(H[i])):
                if logx is True:
                    if labels == None:
                        plt.semilogx(freq[i], tr(H[i][j]))
                    else:
                        plt.semilogx(freq[i], tr(H[i][j]), label=labels[i][j])
                else:
                    if labels == None:
                        plt.plot(freq[i], tr(H[i][j]))
                    else:
                        plt.plot(freq[i], tr(H[i][j]), label=labels[i][j])
        else:
            if logx is True:
                if labels == None:
                    plt.semilogx(freq[i], tr(H[i]))
                else:
                    plt.semilogx(freq[i], tr(H[i]), label=labels[i])
            else:
                if labels == None:
                    plt.plot(freq[i], tr(H[i]))
                else:
                    plt.plot(freq[i], tr(H[i]), label=labels[i])

    # plt.title('Frequency Response Function')
    if labels != None:
        if "loc" in kwargs:
            plt.legend(loc=kwargs["loc"])
        else:
            plt.legend(loc="best")

    if 'xlim' in kwargs:
        plt.xlim(kwargs['xlim'])
    if 'ylim' in kwargs:
        plt.ylim(kwargs['ylim'])
    if 'title' in kwargs:
        plt.title(kwargs['title'])
    if 'xticks' in kwargs:
        xtick_label = []
        for i in range(len(kwargs['xticks'])):
            xtick_label.append(str(kwargs['xticks'][i]))
        plt.xticks(kwargs['xticks'], labels=xtick_label)
    if 'yticks' in kwargs:
        ytick_label = []
        for i in range(len(kwargs['yticks'])):
            ytick_label.append(str(kwargs['yticks'][i]))
        plt.yticks(kwargs['yticks'], labels=ytick_label)
    if 'ylabel' in kwargs:
        plt.ylabel(kwargs['ylabel'])
    else:
        plt.ylabel(tr_str)
    if 'xlabel' in kwargs:
        plt.xlabel(kwargs['xlabel'])
    else:
        plt.xlabel('Frequency [Hz]')

    plt.grid(True, which='both', linestyle='dotted')

    if "save" in kwargs:
        if isinstance(kwargs['save'], str):
            if "dpi" in kwargs:
                plt.savefig(kwargs['save'], dpi=kwargs['dpi'])
            else:
                plt.savefig(kwargs['save'])
        else:
            print("save argument must str")
    plt.tight_layout()
    plt.show()


def directivity(freq, theta, H, transformation='SPL', logx=True, **kwargs):
    """

    :param freq:
    :param theta:
    :param H:
    :param transformation:
    :return:
    """
    from electroacPy.general.gain import dB, SPL

    # associate transformations
    if transformation == "SPL":
        tr = SPL
        tr_str = "SPL [dB]"
        tr_cmap = "turbo"
    elif transformation == "dB":
        tr = dB
        tr_str = "gain [dB]"
        tr_cmap = "turbo"
    elif transformation == "abs":
        tr = np.abs
        tr_str = "magnitude [abs]"
        tr_cmap = "turbo"
    elif transformation == "real":
        tr = np.real
        tr_str = "Real"
        tr_cmap = "seismic"
    elif transformation == "imag":
        tr = np.imag
        tr_str = "Imaginary"
        tr_cmap = "seismic"
    elif transformation == "phase":
        tr = np.angle
        tr_str = "Phase [rad]"
        tr_cmap = "magma"
    else:
        raise Exception("transformation not understood, "
                        "available transformations: 'SPL', 'dB', 'abs', 'phase', 'real', 'imag'")

    if "levels" in kwargs:
        levels = kwargs['levels']
    else:
        levels = 12

    fig, ax = plt.subplots()
    gca = ax.contourf(freq, theta, tr(H), levels, cmap=tr_cmap)
    plt.colorbar(gca, label=tr_str)
    if logx is True:
        ax.set_xscale('log')
    else:
        pass

    if 'xlim' in kwargs:
        plt.xlim(kwargs['xlim'])
    if 'ylim' in kwargs:
        plt.ylim(kwargs['ylim'])
    if 'title' in kwargs:
        plt.title(kwargs['title'])
    if 'xticks' in kwargs:
        xtick_label = []
        for i in range(len(kwargs['xticks'])):
            xtick_label.append(str(kwargs['xticks'][i]))
        plt.xticks(kwargs['xticks'], labels=xtick_label)
    if 'yticks' in kwargs:
        ytick_label = []
        for i in range(len(kwargs['yticks'])):
            ytick_label.append(str(kwargs['yticks'][i]))
        plt.yticks(kwargs['yticks'], labels=ytick_label)

    if 'ylabel' in kwargs:
        ax.set(ylabel=kwargs['ylabel'])
    else:
        ax.set(ylabel="Angle")

    # ax.set(xlabel='Frequency [Hz]', ylabel="Angle")
    ax.set(xlabel="Frequency [Hz]")
    if "save" in kwargs:
        if isinstance(kwargs['save'], str):
            if "dpi" in kwargs:
                plt.savefig(kwargs['save'], dpi=kwargs['dpi'])
            else:
                plt.savefig(kwargs['save'])
        else:
            print("save argument must str")
    plt.show()
    return None
 
def directivityViewer(theta, freq, pMic, xscale='log', fmin=20, fmax=20e3,
                         dBmin=-20, dBmax=False, title="title"):
    """
    Plot the directivity from pMic obtained with getMicPressure for a circular microphone array

    Parameters
    ----------
    theta : narray
        observation angles (in degrees).
    freq : narray
        frequency array.
    pMic : narray
        Pressur obtained with getMicPressure().
    xscale : str, optional
        scaling of the x axis (logarithmic, linear). The default is 'log'.
    fmin : float, optional
        min xlim. The default is 20.
    fmax : float, optional
        max xlim. The default is 20e3.
    dBmin : float, optional
        min ylim. The default is -20.
    dBmax : float, optional
        max ylim. The default is False.
    norm : bool, optional
        set the normalization. The default is True.

    Returns
    -------
    None.

    """
    # compute directivity
    maxAngle = np.array([np.max(abs(pMic), 1)])
    maxMatrix = np.repeat(maxAngle, len(theta), 0).T
    directivity = gtb.gain.dB(np.abs(pMic)/maxMatrix)
    dBmax = int(np.max(gtb.gain.SPL(pMic)))+3
    dBmin = dBmax-40
    
    fig = plt.figure(figsize=(11, 8))
    fig.suptitle(title)
    fig.subplots_adjust(top=0.88)
    ax1 = fig.add_subplot(221)
    gca1 = ax1.contourf(freq, theta, gtb.gain.SPL(pMic).T,
                             np.arange(dBmin, dBmax+3, 3), cmap='turbo')
    ax1.set_xscale('log')
    ax1.set(xlabel="Frequency [Hz]", ylabel="Angle [deg]", xlim=[fmin, fmax], 
            title='Directivity (SPL)')
    plt.colorbar(gca1)
    
    # Total directivity NORMALIZED
    ax2 = fig.add_subplot(223)
    gca2 = ax2.contourf(freq, theta, directivity.T, np.arange(-21, 3, 3), cmap='turbo')
    ax2.set_xscale('log')
    ax2.set(xlabel="Frequency [Hz]", ylabel="Angle [deg]", xlim=[fmin, fmax], 
            title='Normalized directivity')
    plt.colorbar(gca2)
    
    # pressure response
    obsAngle = 0
    ind_angle, value_angle = gtb.findInArray(theta, np.deg2rad(obsAngle))
    ax3 = fig.add_subplot(222)
    ax3.semilogx(freq, gtb.gain.SPL(pMic[:, ind_angle]), label=int(np.rad2deg(value_angle)))
    ax3.set(xlabel='Frequency [Hz]', ylabel='SPL [dB]', ylim=[dBmin, dBmax], 
            title='Pressure response')
    ax3.grid(which='both')
    ax3.legend(loc='best')
    
    # polar
    freqObs = 1000
    ind_freq, value_freq = gtb.findInArray(freq, freqObs)
    ax4 = fig.add_subplot(224, polar=True)
    ax4.plot(np.deg2rad(theta), gtb.gain.SPL(pMic[ind_freq, :]), label=int(value_freq))
    ax4.set(ylim=[dBmin, dBmax], title='Polar directivity')
    ax4.legend(loc='best')

    #defining the cursor
    cursor = Cursor(ax1, horizOn = True, vertOn=True, color='black', linewidth=1.2, 
                    useblit=True)
    
    def onclick(event):
        if event.button is MouseButton.LEFT:
            x = event.xdata
            y = event.ydata
             
            # update subplots -- FRF
            obsAngle = y
            ind_angle, value_angle = gtb.findInArray(theta, obsAngle)
            ax3.semilogx(freq, gtb.gain.SPL(pMic[:, ind_angle]), label=int(value_angle))
            ax3.legend(loc='best')
            
            # update subplots -- polar
            freqObs = x
            ind_freq, value_freq = gtb.findInArray(freq, freqObs)
            ax4.plot(np.deg2rad(theta), 
                     gtb.gain.SPL(pMic[ind_freq, :]), label=int(value_freq))
            ax4.legend(loc='best')
            
            fig.canvas.draw() #redraw the figure
        
        if event.button is MouseButton.RIGHT:
            if len(ax3.lines) > 1:
                ax3.lines[-1].remove()
                ax4.lines[-1].remove()
                ax3.legend(loc='best')
                ax4.legend(loc='best')
                fig.canvas.draw()
        
    fig.canvas.mpl_connect('button_press_event', onclick)
    plt.tight_layout()
    plt.show()
    return cursor
 
    
#%% 3D plots
def bempp_grid(bemOBJ, eval_grid, pmic, radiationCoeff, radiatingElement,
               eval_name, transformation="SPL", export_grid=False):
    """
    Plot grid functions from bempp. "grids" argument takes a list of one or
    multiple grids (evaluation grids), and concatenates everything based on 
    which radiating elements are selected.

    Parameters
    ----------
    bemOBJ : bem object
        bempp simulation,
    eval_grid : list
        eval_grid to plot.
    p_mic : ndarray
        pressure received at microphones (one per eval_grid[i])
    radiationCoeff : ndarray
        coefficients to apply on surfaces
    radiatingElement : list
        radiating elements to sum on system mesh surface.
    eval_name: list of str
        names of eval_grid components
    transformation : str
        scale plot to "SPL" (or "spl"), "real", "imag" or "phase".

    Returns
    -------
    GMSH or Paraview plot.

    """
    import gmsh
    import bempp_cl.api
    import tempfile
    import subprocess

    # get transform
    if transformation in ["SPL", "spl"]:
        T = gtb.gain.SPL
    elif transformation == "real":
        T = np.real
    elif transformation == "imag":
        T = np.imag
    elif transformation == "phase":
        T = np.angle
    
    # get some data from system mesh
    grid         = bemOBJ.grid_init
    Nvert_grid   = grid.vertices.shape[1]
    frequency    = bemOBJ.frequency
    
    # get pressure_grid
    pressure_grid = sumPressureArray(bemOBJ, radiatingElement, radiationCoeff)
    pressure_grid = pressure_grid[:, :Nvert_grid]
          
    pressure = {"system": pressure_grid}
    pressure_id = {"system": 0}
    for i in range(len(eval_name)):
        pressure[eval_name[i]] = pmic[i]
        pressure_id[eval_name[i]] = i+1
        
    # set all grid and plotting grid to a single list (for union!)
    grids = []
    grids.append(bempp_cl.api.Grid(grid.vertices, grid.elements))
    coeff_limits = np.zeros(len(eval_grid)+2, dtype=int)
    coeff_limits[0] = 1
    coeff_limits[1] = Nvert_grid+1
    for i in range(len(eval_name)):
        grid_tmp = bempp_cl.api.Grid(eval_grid[i].vertices, eval_grid[i].elements)
        grids.append(grid_tmp)
        coeff_limits[i+2] = coeff_limits[i+1]+eval_grid[i].vertices.shape[1]

    grid_union = bempp_cl.api.grid.union(grids)
    
    with tempfile.NamedTemporaryFile(delete=False, suffix=".msh") as tmp_file:
        bempp_cl.api.export(tmp_file.name, grid_union, write_binary=False)

    
    # Initialize GMSH
    gmsh.initialize()
    gmsh.option.setNumber("PostProcessing.Format", 5)
    gmsh.option.setNumber("PostProcessing.Link", 1)
    
    # Load total grid and assign view, step, value, etc.
    gmsh.open(tmp_file.name)
    view_tag = gmsh.view.add("evaluation")
    model = gmsh.model.getCurrent()
    for i, surface in enumerate(gmsh.model.getEntities(2)):  # 2 = surfaces
        gmsh.model.addPhysicalGroup(2, [surface[1]], tag=i)
    
    for ps in pressure: 
        ps_id = pressure_id[ps]
        N1 = coeff_limits[ps_id] 
        N2 = coeff_limits[ps_id+1] 
        for j, f in enumerate(frequency):
            data = T(pressure[ps][j, :])
            node_tags = np.arange(N1, N2)
            gmsh.view.addModelData(view_tag, 
                                   j, 
                                   model,
                                   "NodeData", 
                                   node_tags,
                                   data[:, None], 
                                   time=f)
    
    with tempfile.NamedTemporaryFile(delete=False, suffix=".msh") as tmp_export:
        gmsh.write(tmp_export.name)
        gmsh.view.write(view_tag, tmp_export.name)
        
    if export_grid is not False:
        gmsh.write(export_grid)
        gmsh.view.write(view_tag, export_grid)
    
    gmsh.finalize()
    
    # Open the combined file in GMSH viewer
    subprocess.Popen(["gmsh", 
                      "-setnumber", "Mesh.SurfaceEdges", "0", 
                      tmp_export.name], 
                     stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    
    return None

def bempp_grid_mesh(bemOBJ, radiationCoeff, radiatingElement, 
                  transformation="SPL", export_grid=False):
    """
    Plot grid functions from bempp. "grids" argument takes a list of one or
    multiple grids (evaluation grids), and concatenates everything based on 
    which radiating elements are selected.

    Parameters
    ----------
    bemOBJ : bem object
        bempp simulation,
    radiationCoeff : ndarray
        coefficients to apply on surfaces
    radiatingElement : list
        radiating elements to sum on system mesh surface.
    transformation : str
        scale plot to "SPL" (or "spl"), "real", "imag" or "phase".

    Returns
    -------
    GMSH or Paraview plot.

    """
    import gmsh
    import tempfile
    import subprocess

    # get transform
    if transformation in ["SPL", "spl"]:
        T = gtb.gain.SPL
    elif transformation == "real":
        T = np.real
    elif transformation == "imag":
        T = np.imag
    elif transformation == "phase":
        T = np.angle
    
    # get some data from system mesh
    grid         = bemOBJ.grid_init
    Nvert_grid   = grid.vertices.shape[1]
    frequency    = bemOBJ.frequency
    
    # get pressure_grid
    pressure_grid = sumPressureArray(bemOBJ, radiatingElement, radiationCoeff)
    pressure_grid = pressure_grid[:, :Nvert_grid]          
    pressure = {"system": pressure_grid}
        
    # Initialize GMSH
    gmsh.initialize()
    gmsh.option.setNumber("PostProcessing.Format", 5)
    gmsh.option.setNumber("PostProcessing.Link", 1)
    
    # Load total grid and assign view, step, value, etc.
    gmsh.open(bemOBJ.meshPath)
    view_tag = gmsh.view.add("system pressure")
    model = gmsh.model.getCurrent()
    for i, surface in enumerate(gmsh.model.getEntities(2)):
        gmsh.model.addPhysicalGroup(2, [surface[1]], tag=i)
    
    node_tags = np.arange(1, len(bemOBJ.p_mesh[0, 0].coefficients)+1)
    for ps in pressure: 
        for j, f in enumerate(frequency):
            data = T(pressure[ps][j, :])
            gmsh.view.addModelData(view_tag, 
                                   j, 
                                   model,
                                   "NodeData", 
                                   node_tags,
                                   data[:, None], 
                                   time=f)
    
    with tempfile.NamedTemporaryFile(delete=False, suffix=".msh") as tmp_export:
        gmsh.write(tmp_export.name)
        gmsh.view.write(view_tag, tmp_export.name)
        
    if export_grid is not False:
        gmsh.write(export_grid)
        gmsh.view.write(view_tag, export_grid)
    
    gmsh.finalize()
    
    # Open the combined file in GMSH viewer
    subprocess.Popen(["gmsh", 
                      "-setnumber", "Mesh.SurfaceEdges", "0", 
                      tmp_export.name], 
                     stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    
    return None

def pressureField_3D(bemOBJ, xMic, L, W, pMicData, radiatingElement):
    """
    Plot pressure field using PyVista backend.
    :param bemOBJ: object,
        bem object used in the current observation.
    :param L: list,
        list of length of each planar surface defined using "planar" observations.
    :param W: list,
        list of width of each planar surface defined using "planar" observations.
    :param pMicData: ndarray,
        pressure field data.
    :param xMic: ndarray,
        microphones position.
    :param radiatingElement: str, list,
        Index of surfaces currently radiating.
    :return:
    """
    # initialize inputs
    freq_array = bemOBJ.frequency
    meshPlot = []
    
    # put inputs into list (facilitate manipulation when switching between
    # multiple/single plot)
    nObs = len(xMic)
    
    # initialize plotter
    if bemOBJ.domain == "interior":
        pl = pyvista.Plotter(lighting='three lights', shape=(1, 2))
    elif bemOBJ.domain =="exterior":
        pl = pyvista.Plotter(lighting='three lights')
    pl.background_color = 'white'

    for i in range(nObs):
        meshPlot.append(pyvista.StructuredGrid())
        
        # initialisation
        meshPlot[i].points = xMic[i]
        meshPlot[i].points = xMic[i]
        meshPlot[i].dimensions = [len(L[i]), len(W[i]), 1]


    engine = pyvista_spkPlotter(pl, bemOBJ, L, W,
                                pMicData, xMic, radiatingElement, nObs)

    def plotPressurePoint(point):
        position = np.zeros([nObs, 3])
        for i in range(len(xMic)):
            _, position[i, :] = gtb.geometry.findClosestPoint(xMic[i], point[0], point[1], point[2])
        idxp = gtb.geometry.findClosestPoint(position, point[0], point[1], point[2])[0]
        idxm, p = gtb.geometry.findClosestPoint(xMic[idxp], point[0], point[1], point[2])

        pMic_tmp = pMicData[idxp].T
        pMicToPlot = pMic_tmp[idxm, :]
        maxP = np.max(gtb.gain.SPL(pMicToPlot))

        fig, ax = plt.subplots()
        ax.semilogx(bemOBJ.frequency, gtb.gain.SPL(pMicToPlot), label='{}'.format(p))
        ax.grid(which='both')
        ax.set(xlabel='Frequency [Hz]', ylabel='SPL [dB]', ylim=[maxP-40, 
                                                                 maxP+6])
        ax.legend(loc='best')
        plt.show()
        return None


    pl.add_slider_widget(
        callback=lambda value: engine('cfreq', value),
        rng=[np.log10(freq_array[0]), np.log10(freq_array[-1])],
        value=np.log10(250),
        title="Frequency [log]",
        style="modern"
    )

    pl.add_checkbox_button_widget(
        callback=lambda value: engine('real', value),
        value=False,
        position=(5, 100),
    )

    pl.add_checkbox_button_widget(
        callback=lambda value:engine('showMesh', value),
        value=False,
        position=(5, 0),
    )

    pl.add_checkbox_button_widget(
        callback=lambda value:engine('contourP', value),
        value=True,
        position=(5, 50),
    )

    pl.enable_surface_point_picking(
        callback=plotPressurePoint,
        show_message=False,
    )

    obj = pl.show()
    return obj


def boundingBox(bemOBJ, nx, ny, nz, pMicData, xMic, radiatingElement):
    """
    Plot a bounding box observation in a PyVista Plotter
    :param bemOBJ: object,
        bem object used in the current observation.
    :param pMicData: ndarray,
        pressure field data.
    :param xMic: ndarray,
        microphones position.
    :param radiatingElement: str, list,
        Index of surfaces currently radiating.
    :return:
    """
    freq_array = bemOBJ.frequency
    meshPath = bemOBJ.meshPath
    meshPlot = []

    # init plotter
    pl = pyvista.Plotter(lighting='three lights')
    pl.background_color = 'white'
    sizeFactor = bemOBJ.sizeFactor
    vertices = bemOBJ.vertices

    engine = pyvista_boundingBoxPlotter(pl, bemOBJ, nx, ny, nz, pMicData, xMic, radiatingElement)

    pl.add_slider_widget(
        callback=lambda value:engine('cfreq', value),
        rng=[np.log10(freq_array[0]), np.log10(freq_array[-1])],
        value=np.log10(250),
        title="Frequency, log()",
        style="modern"
    )

    pl.add_checkbox_button_widget(
        callback=lambda value:engine('real', value),
        value=False,
        position=(5, 100),
    )

    pl.add_checkbox_button_widget(
        callback=lambda value:engine('showMesh', value),
        value=False,
        position=(5, 0),
    )

    pl.add_checkbox_button_widget(
        callback=lambda value:engine('contourP', value),
        value=True,
        position=(5, 50),
    )

    pl.add_plane_widget(
        callback=lambda normal, origin: engine('normOrigin', (normal, origin)),
        normal=(1, 0, 0),
        origin=(0, 0, 0),
        normal_rotation=True,
    )

    obj = pl.show()
    return obj

def sphericalRadiation(xMic, pMic, freq_array):
    # init plotter
    pl = pyvista.Plotter(lighting='three lights')
    pl.background_color = 'white'

    engine = pyvista_sphericalPlotter(pl, pMic, xMic, freq_array)

    pl.add_slider_widget(
        callback=lambda value:engine('cfreq', value),
        rng=[np.log10(freq_array[0]), np.log10(freq_array[-1])],
        value=np.log10(250),
        title="Frequency, log()",
        style="modern"
    )
    obj = pl.show()
    return obj

## ================================================
# %% PyVista plotter classes (for 3D visualization)
class pyvista_spkPlotter:
    # I think this class and associated methods are more complex than 
    # they should be. For now it works as it is.
    def __init__(self, mesh, bemOBJ, L, W, pMicData, 
                 xMic, radiatingElement, nObs):
        self.output = mesh
        self.bemOBJ = bemOBJ
        self.freq_array = bemOBJ.frequency
        self.vertices = bemOBJ.vertices
        self.sizeFactor = bemOBJ.sizeFactor
        self.L = L
        self.W = W
        self.pMicData = pMicData
        self.xMic = xMic
        self.radiatingElement = radiatingElement
        self.nObs = len(pMicData)

        self.speaker = pyvista.read(bemOBJ.meshPath)
        meshPlot = []
        # pmicdata
        for i in range(nObs):
            meshPlot.append(pyvista.StructuredGrid())
            # initialisation
            meshPlot[i].points = xMic[i]
            meshPlot[i].dimensions = [len(L[i]), len(W[i]), 1]
        self.meshPlot = meshPlot

        # default parameters
        self.kwargs = {
            'real': False,
            'cfreq': '250',
            'showMesh': False,
            'contourP': True
        }

    def __call__(self, param, value):
        self.kwargs[param] = value
        self.update()

    def update(self):
        real = self.kwargs['real']
        cfreq = self.kwargs['cfreq']
        showMesh = self.kwargs['showMesh']
        contourP = self.kwargs['contourP']

        # min / max bounds - title
        maxPressure = []
        minPressure = []
        if real is True:
            for i in range(self.nObs):
                maxPressure.append(np.max(np.real(self.pMicData[i])))
                minPressure.append(np.min(np.real(self.pMicData[i])))
            minPressure = -1 #np.min(minPressure)
            maxPressure = 1  #np.max(maxPressure)
            pressureCoeff_system = sumPressureArray(self.bemOBJ, 
                                                    self.radiatingElement)
            pressureCoeff_system = np.real(pressureCoeff_system)/10
            title_arg = 'Real(Pressure), normalized - '
            cmap_d = 'seismic'
        else:
            for i in range(self.nObs):
                maxPressure.append(np.max(gtb.gain.SPL(self.pMicData[i])))

            maxPressure = np.max(maxPressure) + 5
            minPressure = maxPressure - 75

            # to normalize pressureCoeff_system
            # (pressure on baffle will usually be super high compared to radiated field)
            maxP_abs = 10**(maxPressure/20) * 2e-5

            pressureCoeff_system = sumPressureArray(self.bemOBJ, np.array(self.radiatingElement))
            pressureCoeff_system = gtb.normMinMax(np.abs(pressureCoeff_system), 0, maxP_abs)
            pressureCoeff_system = gtb.gain.SPL(pressureCoeff_system)
            title_arg = 'SPL, dB - '
            cmap_d = 'turbo'

        if contourP is True:
            ncolor=21
        else:
            ncolor=256

        freq_idx, frequency = gtb.findInArray(self.freq_array, 10 ** (cfreq))
        sargs = dict(
            title_font_size=21,
            label_font_size=18,
            shadow=False,
            n_labels=5,
            italic=False,
            font_family="arial",
            color='k',
            title=title_arg + str(round(frequency, 1)) + " Hz",
        )
        
        # system mesh
        try:
            if self.bemOBJ.domain == "exterior":
                self.output.remove_actor('spk_mesh')
                self.output.add_mesh(self.speaker, show_edges=showMesh,
                            scalars=pressureCoeff_system[freq_idx, :self.bemOBJ.spaceP.grid_dof_count // self.sizeFactor],
                            cmap=cmap_d,
                            n_colors=ncolor,
                            show_scalar_bar=False, clim=[minPressure, maxPressure], name='spk_mesh')

            elif self.bemOBJ.domain == "interior":
                self.output.remove_actor('spk_mesh')
                self.output.subplot(0, 0)
                self.output.add_mesh(self.speaker, show_edges=showMesh,
                            scalars=pressureCoeff_system[freq_idx, :self.bemOBJ.spaceP.grid_dof_count // self.sizeFactor],
                            cmap=cmap_d,
                            n_colors=ncolor,
                            show_scalar_bar=False, clim=[minPressure, maxPressure], 
                            name='spk_mesh')
        except:
            pressure_system = pressureCoeff_system[freq_idx, :self.bemOBJ.spaceP.grid_dof_count // self.sizeFactor]
            toAdd = self.vertices - self.bemOBJ.spaceP.grid_dof_count
            scalars_to_plot = np.concatenate((pressure_system, np.zeros(toAdd)))
            if self.bemOBJ.domain == "exterior":
                self.output.remove_actor('spk_mesh')
                self.output.add_mesh(self.speaker, show_edges=showMesh,
                                     scalars=scalars_to_plot,
                                     cmap=cmap_d,
                                     n_colors=ncolor,
                                     show_scalar_bar=False, clim=[minPressure, maxPressure],
                                     name='spk_mesh')
            elif self.bemOBJ.domain == "interior":
                self.output.remove_actor('spk_mesh')
                self.output.subplot(0, 0)
                self.output.add_mesh(self.speaker, show_edges=showMesh,
                            scalars=pressureCoeff_system[freq_idx, :self.bemOBJ.spaceP.grid_dof_count // self.sizeFactor],
                            cmap=cmap_d,
                            n_colors=ncolor,
                            show_scalar_bar=False, 
                            clim=[minPressure, maxPressure], 
                            name='spk_mesh')
        
        show_sc_b = []
        for i in range(self.nObs):
            if i == 0:
                show_sc_b.append(True)
            else:
                show_sc_b.append(False)

            data = self.pMicData[i].T
            pMicRes = np.reshape(data[:, int(freq_idx)],
                                 [len(self.L[i]), len(self.W[i])])

            # def real or SPL pressure plot
            if real is True:
                # pMicToPlot = gtb.normMinMax(np.real(pMicRes), minPressure, maxPressure)
                pMicToPlot = np.real(pMicRes)
            else:
                pMicToPlot = gtb.gain.SPL(pMicRes)

            # observation mesh
            if self.bemOBJ.domain == "exterior":
                self.output.remove_actor('SPL_{}'.format(i))
                self.output.add_mesh(self.meshPlot[i], show_edges=showMesh, scalars=pMicToPlot, cmap=cmap_d,
                            scalar_bar_args=sargs, clim=[minPressure, maxPressure], n_colors=ncolor,
                            name='SPL_{}'.format(i), show_scalar_bar=show_sc_b[i])
            elif self.bemOBJ.domain == "interior":
                self.output.subplot(0, 1)
                self.output.remove_actor('SPL_{}'.format(i))
                self.output.add_mesh(self.meshPlot[i], show_edges=showMesh, scalars=pMicToPlot, cmap=cmap_d,
                            scalar_bar_args=sargs, clim=[minPressure, maxPressure], n_colors=ncolor,
                            name='SPL_{}'.format(i), show_scalar_bar=show_sc_b[i])
                self.output.add_mesh(self.speaker, show_scalar_bar=False, color='grey',
                                     opacity=0.15, name='boundary_{}'.format(i))
                self.output.link_views()
        
        if self.bemOBJ.domain == "exterior": # if used for interior plot, crashes
            self.output.add_axes(color='k') # this mf crashes
        
        _ = self.output.show_grid(color='k') # this one does not
        return


class pyvista_boundingBoxPlotter:
    def __init__(self, mesh, bemOBJ, nx, ny, nz, pMicData, xMic, radiatingElement):
        self.output = mesh
        self.bemOBJ = bemOBJ
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.freq_array = bemOBJ.frequency
        self.vertices = bemOBJ.vertices
        self.sizeFactor = bemOBJ.sizeFactor
        self.pMicData = pMicData
        self.xMic = xMic
        self.radiatingElement = radiatingElement

        self.speaker = pyvista.read(bemOBJ.meshPath) # will display the speaker's mesh

        # pmicdata
        meshPlot = pyvista.StructuredGrid()
        meshPlot.points = xMic
        meshPlot.dimensions = [nx, ny, nz]
        self.meshPlot = meshPlot   # will display bounding box mesh (unstructured grid?)

        # default parameters
        self.kwargs = {
            'real': False,
            'cfreq': '250',
            'showMesh': False,
            'contourP': True,
            'normal': (1, 0, 0),
            'origin': (0, 0, 0)
        }

    def __call__(self, param, value):
        self.kwargs[param] = value
        self.update()

    def update(self):
        bemOBJ = self.bemOBJ
        real = self.kwargs['real']
        cfreq = self.kwargs['cfreq']
        showMesh = self.kwargs['showMesh']
        contourP = self.kwargs['contourP']
        # normOrigin = self.kwargs['normOrigin']
        normal = self.kwargs['normal']
        origin = self.kwargs['origin']

    
        # normal = normOrigin[0]
        # origin = normOrigin[1]

        # min / max bounds - title
        if real is True:
            minPressure = np.min(np.real(self.pMicData))
            maxPressure = np.max(np.real(self.pMicData))
            pressureCoeff_system = sumPressureArray(self.bemOBJ, self.radiatingElement)
            pressureCoeff_system = gtb.normMinMax(np.real(pressureCoeff_system), minPressure, maxPressure)
            title_arg = 'Real(Pressure), Pa - '
            cmap_d = 'seismic'
        else:
            maxPressure = np.max(gtb.gain.SPL(self.pMicData)) + 5
            minPressure = maxPressure - 75

            # to normalize pressureCoeff_system
            # (pressure on baffle will usually be super high compared to radiated field)
            maxP_abs = 10 ** (maxPressure / 20) * 2e-5

            pressureCoeff_system = sumPressureArray(self.bemOBJ, self.radiatingElement)
            pressureCoeff_system = gtb.normMinMax(np.abs(pressureCoeff_system), 0, maxP_abs)
            pressureCoeff_system = gtb.gain.SPL(pressureCoeff_system)
            title_arg = 'SPL, dB - '
            cmap_d = 'turbo'

        if contourP is True:
            ncolor = 21
        else:
            ncolor = 256

        freq_idx, frequency = gtb.findInArray(self.freq_array, 10 ** (cfreq))
        sargs = dict(
            title_font_size=21,
            label_font_size=18,
            shadow=False,
            n_labels=5,
            italic=False,
            font_family="arial",
            color='k',
            title=title_arg + str(round(frequency, 1)) + " Hz",
        )


        data = self.pMicData.T
        pMicRes = data[:, int(freq_idx)]

        # def real or SPL pressure plot
        if real is True:
            pMicToPlot = gtb.normMinMax(np.real(pMicRes), minPressure, maxPressure)
        else:
            pMicToPlot = gtb.gain.SPL(pMicRes)

        # system mesh
        _ = self.output.remove_actor('spk_mesh')
        self.output.add_mesh(self.speaker, show_edges=showMesh,
                             scalars=pressureCoeff_system[freq_idx, :self.vertices // self.sizeFactor],
                             cmap=cmap_d, n_colors=ncolor, show_scalar_bar=False,
                             clim=[minPressure, maxPressure], name='spk_mesh')

        # observation mesh
        meshPlot = self.meshPlot
        meshPlot.point_data['scalars'] = pMicToPlot
        slc = meshPlot.slice(normal=normal, origin=origin)
        # self.output.add_mesh(meshPlot.outline(), color='k')
        self.output.add_mesh(slc, show_edges=showMesh, show_scalar_bar=True,
                                   cmap=cmap_d,  clim=[minPressure, maxPressure], n_colors=ncolor,
                                   name='SPL_slice', scalar_bar_args=sargs)
        _ = self.output.show_grid(color='k')
        self.output.add_axes(color='k')
        return


class pyvista_sphericalPlotter:
    def __init__(self, mesh, pMic, xMic, frequency):
        self.output = mesh
        self.pMic = pMic
        self.xMic = xMic
        self.frequency = frequency

        ELEVATION = np.arccos(xMic[:, 2] / np.sqrt(xMic[:, 0] ** 2 + xMic[:, 1] ** 2 + xMic[:, 2] ** 2))
        AZIMUTH = np.zeros(len(xMic))
        self.R = np.zeros([len(frequency), len(xMic)])
        for i in range(len(xMic)):
            if xMic[i, 1] > 0:
                AZIMUTH[i] = np.arccos(xMic[i, 0] / np.sqrt(xMic[i, 0] ** 2 + xMic[i, 1] ** 2))
            elif xMic[i, 1] < 0:
                AZIMUTH[i] = - np.arccos(xMic[i, 0] / np.sqrt(xMic[i, 0] ** 2 + xMic[i, 1] ** 2))

        self.elevation = ELEVATION
        self.azimuth = AZIMUTH

        self.kwargs = {
            'cfreq': '250',
        }

    def __call__(self, param, value):
        self.kwargs[param] = value
        self.update()

    def update(self):
        cfreq = self.kwargs['cfreq']
        # print(cfreq)
        R = self.R
        elevation = self.elevation
        azimuth = self.azimuth

        freq_idx, frequency = gtb.findInArray(self.frequency, 10 ** (cfreq))

        sargs = dict(
            title_font_size=21,
            label_font_size=18,
            shadow=False,
            n_labels=5,
            italic=False,
            font_family="arial",
            color='k',
            title="SPL, dB - " + str(round(frequency, 1)) + " Hz",
        )

        cloud = pyvista.PolyData(self.xMic)
        volume = cloud.delaunay_3d(alpha=2.)
        shell = volume.extract_geometry()
        shell.points /= np.sqrt(self.xMic[0, 0]**2 + self.xMic[0, 1]**2 + self.xMic[0, 2]**2)
        shell.points *= np.tile(gtb.gain.SPL(self.pMic[freq_idx, :]), (3, 1)).T

        _ = self.output.remove_actor('balloon')
        self.output.add_mesh(shell, name='balloon', cmap="turbo",
                             scalars=gtb.gain.SPL(self.pMic[freq_idx, :]),
                             clim=[np.max(gtb.gain.SPL(self.pMic))-30, np.max(gtb.gain.SPL(self.pMic)) + 6],
                             n_colors=20, scalar_bar_args=sargs)
        _ = self.output.show_grid(color='k')
        self.output.add_axes(color='k')
        return None





"""Functions for creating boundary integral operators.

Documentation of the functionality of this module and examples
can be found at https://bempp.com/handbook/api/boundary_operators.html.
"""

from . import laplace
from . import helmholtz
from . import modified_helmholtz
from . import maxwell
from . import sparse



"""Data type checks."""

import numpy as _np


def check_type(name, default="float64"):
    """Try to convert input into a numpy.dtype object."""
    value = str(name) if name is not None else default
    if value not in ["float32", "float64", "complex64", "complex128"]:
        raise ValueError("Incorrect type (%s)" % value)

    return _np.dtype(value)


def combined_type(dtype1, dtype2):
    """Return a type that is compatible with dtype1 and dtype2."""
    new_d1 = check_type(dtype1)
    new_d2 = check_type(dtype2)

    return (new_d1.type(1) * new_d2.type(1)).dtype
